from .. import Secret


class OpenAISecret(Secret):
    """
    .. note::
            To create an OpenAISecret, please use the factory method :func:`secret` with ``provider="openai"``.
    """

    _PROVIDER = "openai"
    _DEFAULT_ENV_VARS = {"api_key": "OPENAI_API_KEY"}



from typing import Dict, Optional

from kubetorch.globals import config

from .secret import Secret


def secret(
    name: Optional[str] = None,
    provider: Optional[str] = None,
    path: Optional[str] = None,
    env_vars: Optional[Dict] = None,
    namespace: Optional[str] = config.namespace,
    override: Optional[bool] = False,
) -> Secret:
    """
    Builds an instance of :class:`Secret`. At most one of `provider`, `path`, or `env_vars` can be provided, to maintain
    one source of truth. For a provider, the values are inferred from the default path or environment variables for that
    provider. To load a secret by name, provide its name and namespace.

    Args:
        namespace (str, optional): Namespace to load the secret from, if we create a secret from name. Default: "default".
        name (str, optional): Name to assign the resource. If none is provided, resource name defaults to the
            provider name.
        provider (str, optional): Provider corresponding to the secret (e.g. "aws", "gcp"). To see all supported provider
            types, run ``kt.Secret.builtin_providers(as_str=True)``.
        path (str, optional): Path where the secret values are held.
        env_vars (Dict, optional): Dictionary mapping secret keys to the corresponding
            environment variable key.
        override (Bool, optional): If True, override the secret's values in Kubernetes if a secret with the same name already exists.

    Returns:
        Secret: The resulting secret object.

    Examples:

    .. code-block:: python

        import kubetorch as kt

        local_secret = kt.secret(name="in_memory_secret", values={"secret_key": "secret_val"})
        aws_secret = kt.secret(provider="aws")
        gcp_secret = kt.secret(name="my-gcp-secret", path="~/.gcp/credentials")
        lambda_secret = kt.secret(name= "my-lambda-secret", env_vars={"api_key": "LAMBDA_API_KEY"})
    """

    # env_vars or path or provider are provided
    valid_input = sum([bool(x) for x in [provider, path, env_vars]]) == 1 or (provider and path)
    valid_from_name_input = sum([bool(x) for x in [provider, path, env_vars]]) == 0 and name

    if not (valid_from_name_input or valid_input):
        raise ValueError(
            "You must provide exactly one of: `provider`, `path`, or `env_vars`. Alternatively, you may provide `name` to load a secret from name."
        )

    if valid_input:
        if provider:
            return Secret.from_provider(provider=provider, name=name, path=path, override=override)
        elif path and not provider:  # the case where provider + path are provided are
            return Secret.from_path(path=path, name=name, override=override)
        elif env_vars:
            return Secret.from_env(env_vars=env_vars, name=name, override=override)
    else:
        return Secret.from_name(name=name, namespace=namespace)



import os
from functools import update_wrapper
from typing import List, Union

from kubetorch.resources.callables.cls.cls import cls
from kubetorch.resources.callables.fn.fn import fn
from kubetorch.resources.compute.compute import Compute

# Helper class which allows us to chain decorators in a way that allows us to reverse the order of the decorators
# The `compute` decorator ultimately unwinds the calls to properly construct the Module.
class PartialModule:
    def __init__(
        self,
        fn_or_cls=None,
        distribute_args=None,
        autoscale_args=None,
        async_=False,
    ):
        self.fn_or_cls = fn_or_cls
        self.distribute_args = distribute_args
        self.autoscale_args = autoscale_args
        self.async_ = async_


# @kubetorch.compute decorator that the user can use to wrap a function they want to deploy to a cluster,
# and then deploy it with `kt deploy my_app.py` (we collect all the decorated functions imported in the file
# to deploy them).
def compute(get_if_exists: bool = False, reload_prefixes: Union[str, List[str]] = [], **kwargs):
    def decorator(func_or_cls):
        from kubetorch.globals import disable_decorators

        if disable_decorators:
            return func_or_cls

        if isinstance(func_or_cls, PartialModule):
            distribute_args = func_or_cls.distribute_args
            autoscale_args = func_or_cls.autoscale_args
            async_ = func_or_cls.async_
            func_or_cls = func_or_cls.fn_or_cls
        else:
            distribute_args = None
            autoscale_args = None
            async_ = False

        # If we're on the server attempting to load this function or class, just return it as is
        if (
            os.environ.get("KT_CLS_OR_FN_NAME") == func_or_cls.__name__
            and os.environ.get("KT_MODULE_NAME") == func_or_cls.__module__
        ):
            return func_or_cls

        module_name = kwargs.pop("name", None)
        kt_deploy_mode = os.environ.get("KT_CLI_DEPLOY_MODE") == "1"

        if isinstance(func_or_cls, type):
            new_module = cls(
                func_or_cls,
                name=module_name,
                get_if_exists=get_if_exists,
                reload_prefixes=reload_prefixes,
            )
        else:
            new_module = fn(
                func_or_cls,
                name=module_name,
                get_if_exists=get_if_exists,
                reload_prefixes=reload_prefixes,
            )

        if async_:
            new_module.async_ = async_

        if kt_deploy_mode:
            # Create new Compute and pass in remaining kwargs only in kt deploy mode, not when importing
            # Imported kt module will be reloaded from name when called
            new_module.compute = Compute(**kwargs)
            new_module.compute.service_name = new_module.service_name
            if distribute_args:
                distribute_args, distribute_kwargs = distribute_args
                new_module.compute.distribute(*distribute_args, **distribute_kwargs)
            if autoscale_args:
                autoscale_args, autoscale_kwargs = autoscale_args
                new_module.compute.autoscale(*autoscale_args, **autoscale_kwargs)

        # update_wrapper(new_module, func_or_cls)
        return new_module

    return decorator


def async_(func_or_cls):
    from kubetorch.globals import disable_decorators

    if disable_decorators:
        return func_or_cls

    # If it's already a PartialModule (from other decorators), update it
    if isinstance(func_or_cls, PartialModule):
        func_or_cls.async_ = True
        return func_or_cls

    # Otherwise, create a new PartialModule
    partial_module = PartialModule(fn_or_cls=func_or_cls, async_=True)
    update_wrapper(partial_module, func_or_cls)
    return partial_module


def distribute(*args, **kwargs):
    def decorator(func_or_cls):
        from kubetorch.globals import disable_decorators

        if disable_decorators:
            return func_or_cls

        # This is a partial so the order of decorator chaining can be reversed for best aesthetics
        # the deploy method will actually call .distribute on the function or class after it's been deployed
        partial_module = PartialModule(fn_or_cls=func_or_cls, distribute_args=(args, kwargs))
        update_wrapper(partial_module, func_or_cls)
        return partial_module

    return decorator


def autoscale(*args, **kwargs):
    def decorator(func_or_cls):
        from kubetorch.globals import disable_decorators

        if disable_decorators:
            return func_or_cls

        # This is a partial so the order of decorator chaining can be reversed for best aesthetics
        # the deploy method will actually call .distribute on the function or class after it's been deployed
        partial_module = PartialModule(fn_or_cls=func_or_cls, autoscale_args=(args, kwargs))
        update_wrapper(partial_module, func_or_cls)
        return partial_module

    return decorator



import asyncio
import base64
import hashlib
import json
import os
import signal

import subprocess
import threading
import time
import urllib.parse
from concurrent.futures import ThreadPoolExecutor, TimeoutError
from contextlib import contextmanager
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from typing import List, Optional

import httpx
import typer
import yaml
from kubernetes import client
from kubernetes.client.rest import ApiException
from pydantic import BaseModel
from rich import box
from rich.console import Console
from rich.style import Style
from rich.table import Table
from websocket import create_connection

import kubetorch.serving.constants as serving_constants

from kubetorch import globals
from kubetorch.config import KubetorchConfig
from kubetorch.constants import MAX_PORT_TRIES

from kubetorch.resources.compute.utils import is_port_available
from kubetorch.servers.http.utils import stream_logs_websocket_helper, StreamType
from kubetorch.serving.utils import wait_for_port_forward
from kubetorch.utils import load_kubeconfig

from .constants import BULLET_UNICODE, CPU_RATE, DOUBLE_SPACE_UNICODE, GPU_RATE

from .logger import get_logger

console = Console()

logger = get_logger(__name__)


# ------------------ Billing helpers--------------------
class UsageData(BaseModel):
    date_start: str
    date_end: str
    cpu_hours: float
    gpu_hours: float


class BillingTotals(BaseModel):
    cpu: float
    gpu: float


class BillingCosts(BaseModel):
    cpu: float
    gpu: float


class BillingRequest(BaseModel):
    license_key: str
    signature: str
    file_name: str
    username: Optional[str] = None
    usage_data: UsageData
    totals: BillingTotals
    costs: BillingCosts


# ------------------ Generic helpers--------------------
class VolumeAction(str, Enum):
    list = "list"
    create = "create"
    delete = "delete"
    ssh = "ssh"


class SecretAction(str, Enum):
    list = "list"
    create = "create"
    delete = "delete"
    describe = "describe"


def default_typer_values(*args):
    """Convert typer model arguments to their default values or types, so the CLI commands can be also imported and
    called in Python if desired."""
    new_args = []
    for arg in args:
        if isinstance(arg, typer.models.OptionInfo):
            # Replace the typer model with its value
            arg = arg.default if arg.default is not None else arg.type
        elif isinstance(arg, typer.models.ArgumentInfo):
            # Replace the typer model with its value
            arg = arg.default if arg.default is not None else arg.type
        new_args.append(arg)
    return new_args


def validate_config_key(key: str = None):
    if key is None:
        return

    valid_keys = {name for name, attr in vars(KubetorchConfig).items() if isinstance(attr, property)}
    if key not in valid_keys:
        raise typer.BadParameter(f"Valid keys are: {', '.join(sorted(valid_keys))}")
    return key


def get_pods_for_service_cli(name: str, namespace: str, core_api):
    """Get pods for a service using unified label selector."""
    # Use unified service label - works for all deployment modes
    label_selector = f"kubetorch.com/service={name}"
    return core_api.list_namespaced_pod(
        namespace=namespace,
        label_selector=label_selector,
    )


def service_name_argument(*args, required: bool = True, **kwargs):
    def _lowercase(value: str) -> str:
        return value.lower() if value else value

    default = ... if required else ""
    return typer.Argument(default, callback=_lowercase, *args, **kwargs)


def get_deployment_mode(name: str, namespace: str, custom_api, apps_v1_api) -> str:
    """Validate service exists and return deployment mode."""
    try:
        original_name = name
        deployment_mode = detect_deployment_mode(name, namespace, custom_api, apps_v1_api)
        # If service not found and not already prefixed with username, try with username prefix
        if not deployment_mode and globals.config.username and not name.startswith(globals.config.username + "-"):
            name = f"{globals.config.username}-{name}"
            deployment_mode = detect_deployment_mode(name, namespace, custom_api, apps_v1_api)

        if not deployment_mode:
            console.print(f"[red]Failed to load service [bold]{original_name}[/bold] in namespace {namespace}[/red]")
            raise typer.Exit(1)
        console.print(f"Found [green]{deployment_mode}[/green] service [blue]{name}[/blue]")
        return name, deployment_mode

    except ApiException as e:
        console.print(f"[red]Kubernetes API error: {e}[/red]")
        raise typer.Exit(1)


def validate_pods_exist(name: str, namespace: str, core_api) -> list:
    """Validate pods exist for service and return pod list."""
    pods = get_pods_for_service_cli(name, namespace, core_api)
    if not pods.items:
        console.print(f"\n[red]No pods found for service {name} in namespace {namespace}[/red]")
        console.print(f"You can view the service's status using:\n [yellow]  kt status {name}[/yellow]")
        raise typer.Exit(1)
    return pods.items


@contextmanager
def port_forward_to_pod(
    pod_name,
    namespace: str = None,
    local_port: int = 8080,
    remote_port: int = serving_constants.DEFAULT_NGINX_PORT,
    health_endpoint: str = None,
):

    load_kubeconfig()
    for attempt in range(MAX_PORT_TRIES):
        candidate_port = local_port + attempt
        if not is_port_available(candidate_port):
            logger.debug(f"Local port {candidate_port} is already in use, trying again...")
            continue

        cmd = [
            "kubectl",
            "port-forward",
            f"pod/{pod_name}",
            f"{candidate_port}:{remote_port}",
            "--namespace",
            namespace,
        ]
        logger.debug(f"Running port-forward command: {' '.join(cmd)}")

        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, start_new_session=True)

        try:
            wait_for_port_forward(
                process,
                candidate_port,
                health_endpoint=health_endpoint,
                validate_kubetorch_versions=False,
            )
            time.sleep(2)
            yield candidate_port
            return

        finally:
            if process:
                try:
                    os.killpg(os.getpgid(process.pid), signal.SIGTERM)
                    process.wait()
                except (ProcessLookupError, OSError):
                    # Process may have already terminated
                    pass

    raise RuntimeError(f"Could not bind available port after {MAX_PORT_TRIES} attempts")


def get_last_updated(pod):
    conditions = pod["status"].get("conditions", [])
    latest = max(
        (c.get("lastTransitionTime") for c in conditions if c.get("lastTransitionTime")),
        default="",
    )
    return latest


# ------------------ Reporting helpers--------------------
def upload_report(
    usage_data: dict,
    signature: str,
    costs: BillingCosts,
    totals: BillingTotals,
    file_name: str,
    license_key: str,
    username: str = None,
):
    billing_request = BillingRequest(
        license_key=license_key,
        signature=signature,
        file_name=file_name,
        username=username,
        usage_data=UsageData(**usage_data),
        costs=costs,
        totals=totals,
    )

    url = "https://auth.run.house/v1/billing/report"
    resp = httpx.post(url, json=billing_request.model_dump())
    if resp.status_code != 200:
        console.print("[red]Failed to send billing report[/red]")
        raise typer.Exit(1)


def export_report_pdf(report_data, filename):
    try:
        from reportlab.lib import colors
        from reportlab.lib.pagesizes import letter
        from reportlab.pdfgen import canvas
    except ImportError:
        console.print(
            "[red]ReportLab is required for downloading the report. Please install it "
            "with `pip install reportlab`.[/red]"
        )
        raise typer.Exit(1)

    usage_data: dict = report_data["usage_report"]
    report_str = json.dumps(report_data, sort_keys=True)
    signature = base64.b64encode(hashlib.sha256(report_str.encode()).digest()).decode()

    c = canvas.Canvas(filename, pagesize=letter)
    width, height = letter

    # Sidebar
    sidebar_color = colors.HexColor("#4B9CD3")
    c.setFillColor(sidebar_color)
    c.roundRect(0, 0, 18, height, 0, fill=1, stroke=0)
    c.setFillColor(colors.black)

    y = height - 60

    # Header Title
    c.setFont("Helvetica-Bold", 26)
    c.setFillColor(sidebar_color)
    c.drawCentredString(width / 2, y, "Kubetorch Usage Report")
    c.setFillColor(colors.black)
    y -= 30

    # Header Bar
    c.setStrokeColor(sidebar_color)
    c.setLineWidth(2)
    c.line(40, y, width - 40, y)
    y -= 20

    # Info Box
    c.setFillColor(colors.whitesmoke)
    c.roundRect(40, y - 60, width - 80, 60, 8, fill=1, stroke=0)
    c.setFillColor(colors.black)
    c.setFont("Helvetica-Bold", 12)
    c.drawString(55, y - 20, "Username:")
    c.drawString(55, y - 35, "Cluster:")
    c.setFont("Helvetica", 12)
    c.drawString(130, y - 20, report_data["username"])
    c.drawString(130, y - 35, report_data.get("cluster_name", "N/A"))
    y -= 100

    # Usage Summary Section
    c.setFont("Helvetica-Bold", 15)
    c.setFillColor(sidebar_color)
    c.drawString(40, y, "Usage Summary")
    c.setFillColor(colors.black)
    y -= 25

    # Table Outline (dashed)
    table_left = 40
    table_width = width - 80
    row_height = 18
    num_rows = 2  # header + data
    table_height = row_height * num_rows

    # Table Header (centered text)
    header_height = row_height
    c.setFillColor(sidebar_color)
    c.roundRect(table_left, y - header_height, table_width, header_height, 4, fill=1, stroke=0)
    c.setFont("Helvetica-Bold", 11)
    c.setFillColor(colors.white)
    header_y = y - header_height + 5
    c.drawString(table_left + 10, header_y, "Start Date")
    c.drawString(table_left + 90, header_y, "End Date")
    c.drawString(table_left + 200, header_y, "vCPU Hours")
    c.drawString(table_left + 300, header_y, "GPU Hours")
    c.setFillColor(colors.black)

    # Dashed outline (starts at header, not above)
    c.setStrokeColor(sidebar_color)
    c.setDash(4, 4)
    c.roundRect(table_left, y - table_height, table_width, table_height, 6, fill=0, stroke=1)
    c.setDash()
    y -= header_height

    # Table Rows
    c.setFont("Helvetica", 10)
    y -= row_height
    c.drawString(table_left + 10, y + 5, usage_data["date_start"])
    c.drawString(table_left + 90, y + 5, usage_data["date_end"])
    c.drawRightString(table_left + 270, y + 5, f"{usage_data['cpu_hours']:.2f}")
    c.drawRightString(table_left + 370, y + 5, f"{usage_data['gpu_hours']:.2f}")

    y -= 30

    # Invoice Calculation
    total_cpu = usage_data["cpu_hours"]
    total_gpu = usage_data["gpu_hours"]
    cpu_cost = total_cpu * CPU_RATE
    gpu_cost = total_gpu * GPU_RATE
    total_cost = cpu_cost + gpu_cost

    y -= 20
    c.setFont("Helvetica-Bold", 13)
    c.setFillColor(sidebar_color)
    c.drawString(40, y, "Invoice Summary")
    c.setFillColor(colors.black)
    y -= 18

    c.setFont("Helvetica", 11)
    c.drawString(50, y, f"Total vCPU Hours: {total_cpu:.2f} @ ${CPU_RATE:.2f}/hr")
    c.drawRightString(width - 50, y, f"${cpu_cost:.2f}")
    y -= 15
    c.drawString(50, y, f"Total GPU Hours: {total_gpu:} @ ${GPU_RATE:.2f}/hr")
    c.drawRightString(width - 50, y, f"${gpu_cost:.2f}")
    y -= 15

    line_left = 50
    line_right = width - 50
    c.setStrokeColor(sidebar_color)
    c.setLineWidth(1.5)
    c.line(line_left, y, line_right, y)
    y -= 15

    c.setFont("Helvetica-Bold", 12)
    c.drawString(50, y, "Total Due:")
    c.setFont("Helvetica-Bold", 12)
    c.setFillColor(colors.HexColor("#008000"))
    c.drawRightString(width - 50, y, f"${total_cost:.2f}")
    c.setFillColor(colors.black)
    y -= 30

    # Signature and footer
    sig_y = 80
    sig_val_y = sig_y - 15
    footer_y = sig_val_y - 40

    # Signature at the bottom
    c.setFont("Helvetica-Bold", 12)
    c.setFillColor(colors.black)
    c.drawString(40, sig_y, "Signature:")
    c.setFont("Courier-Oblique", 8)
    c.setFillColor(colors.HexColor("#888888"))
    c.drawString(40, sig_val_y, signature)
    c.setFillColor(colors.black)

    # Footer
    c.setFont("Helvetica-Oblique", 8)
    c.setFillColor(colors.HexColor("#888888"))
    c.drawString(40, footer_y, f"Generated on {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")
    c.setFillColor(colors.black)

    c.save()
    return signature


def print_usage_table(usage_data, cluster_name):
    table = Table(title="Usage Summary")
    table.add_column("Start Date")
    table.add_column("End Date")
    table.add_column("vCPU Hours")
    table.add_column("GPU Hours")
    table.add_row(
        usage_data["date_start"],
        usage_data["date_end"],
        str(usage_data["cpu_hours"]),
        str(usage_data["gpu_hours"]),
    )
    console.print(table)
    console.print(f"[dim]Cluster: {str(cluster_name)}[/dim]")


def get_last_n_calendar_weeks(n_weeks):
    """Return a list of (week_start, week_end) tuples for the last n full calendar weeks (Mon–Sun),
    not including this week."""
    today = datetime.utcnow().date()
    # Find the most recent Monday before today (not including today if today is Monday)
    if today.weekday() == 0:
        last_monday = today - timedelta(days=7)
    else:
        last_monday = today - timedelta(days=today.weekday())

    weeks = []
    for i in range(n_weeks):
        week_start = last_monday - timedelta(weeks=i)
        week_end = week_start + timedelta(days=6)  # Monday + 6 = Sunday
        weeks.append((week_start, week_end))

    weeks.reverse()  # So oldest week is first
    return weeks


def get_usage_data(prom, weeks):
    from datetime import datetime, timedelta

    days = weeks * 7
    end_time = datetime.now()
    start_time = end_time - timedelta(days=days)

    # sum of CPU-seconds used by all cores for that container (ex: 2 cores for 1 second = 2 seconds)
    cpu_query = f'increase(container_cpu_usage_seconds_total{{container="kubetorch"}}[{days}d])'
    cpu_result = prom.custom_query(cpu_query)

    total_cpu_seconds = 0
    if cpu_result:
        for series in cpu_result:
            cpu_val = float(series["value"][1])
            total_cpu_seconds += cpu_val

    cpu_hours = round(total_cpu_seconds / 3600, 2)

    # requested GPUs × time they were running
    gpu_query = f'sum_over_time(kube_pod_container_resource_requests{{resource="nvidia_com_gpu", container="kubetorch"}}[{days}d])'
    gpu_result = prom.custom_query(gpu_query)

    # Convert to "GPU hours" over the period
    total_gpu_seconds = sum(float(s["value"][1]) for s in gpu_result or [])
    gpu_hours = total_gpu_seconds / 3600

    usage = {
        "date_start": start_time.strftime("%Y-%m-%d"),
        "date_end": end_time.strftime("%Y-%m-%d"),
        "cpu_hours": round(cpu_hours, 2),
        "gpu_hours": round(gpu_hours, 2),
    }

    return usage


# ------------------ Monitoring helpers--------------------
def get_service_metrics(prom, pod_name: str, pod_node: str, running_on_gpu: bool):
    """Get CPU, GPU (if relevant) and memory metrics for a pod"""

    def extract_prometheus_metric_value(query) -> float:
        result = prom.custom_query(query=query)
        return float(result[0].get("value")[1]) if result else 0

    # --- CPU metrics --- #
    cpu_query_time_window = "30s"
    used_cpu_query = (
        f"sum(rate(container_cpu_usage_seconds_total{{pod='{pod_name}', "
        f"container='kubetorch'}}[{cpu_query_time_window}]))"
    )
    requested_cpu_query = (
        f"sum(kube_pod_container_resource_requests{{pod='{pod_name}', " f"resource='cpu', container='kubetorch'}})"
    )

    used_cpu_result = extract_prometheus_metric_value(used_cpu_query)

    requested_cpu_result = extract_prometheus_metric_value(requested_cpu_query)

    cpu_util = (
        round((100 * (used_cpu_result / requested_cpu_result)), 3) if used_cpu_result and requested_cpu_result else 0
    )

    memory_usage_query = f"container_memory_usage_bytes{{pod='{pod_name}', container='kubetorch'}} / 1073741824"
    memory_usage = round(extract_prometheus_metric_value(memory_usage_query), 3)

    machine_mem_query = f"machine_memory_bytes{{node='{pod_node}'}} / 1073741824"  # convert to GB
    machine_mem_result = extract_prometheus_metric_value(machine_mem_query)

    cpu_mem_percent = round((memory_usage / machine_mem_result) * 100, 3) if machine_mem_result else 0
    collected_metrics = {
        "cpu_util": cpu_util,
        "used_cpu": round(used_cpu_result, 4),
        "requested_cpu": round(requested_cpu_result, 4),
        "cpu_memory_usage": memory_usage,
        "cpu_memory_total": round(machine_mem_result, 3),
        "cpu_memory_usage_percent": cpu_mem_percent,
    }

    # --- GPU metrics --- #
    if running_on_gpu:
        gpu_util_query = f"DCGM_FI_DEV_GPU_UTIL{{exported_pod='{pod_name}', exported_container='kubetorch'}}"
        gpu_mem_used_query = (
            f"DCGM_FI_DEV_FB_USED{{exported_pod='{pod_name}', " f"exported_container='kubetorch'}} * 1.048576 / 1000"
        )  # convert MiB to MB to GB
        gpu_mem_free_query = (
            f"DCGM_FI_DEV_FB_FREE{{exported_pod='{pod_name}', " f"exported_container='kubetorch'}} * 1.048576 / 1000"
        )  # convert MiB to MB to GB

        gpu_util = extract_prometheus_metric_value(gpu_util_query)
        gpu_mem_used = round(extract_prometheus_metric_value(gpu_mem_used_query), 3)
        gpu_mem_free = extract_prometheus_metric_value(gpu_mem_free_query)
        gpu_mem_total = gpu_mem_free + gpu_mem_used
        gpu_mem_percent = (
            round(100 * (gpu_mem_used / gpu_mem_total), 2) if gpu_mem_used else 0
        )  # raw approximation, because total_allocated_gpu_memory is not collected

        gpu_metrics = {
            "gpu_util": gpu_util,
            "gpu_memory_usage": gpu_mem_used,
            "gpu_memory_total": round(gpu_mem_total, 3),
            "gpu_memory_usage_percent": gpu_mem_percent,
        }

        collected_metrics.update(gpu_metrics)

    return collected_metrics


def get_current_cluster_name():
    try:
        from kubernetes import config as k8s_config

        k8s_config.load_incluster_config()
        # In-cluster: return a generic name or the service host
        return os.environ.get("CLUSTER_NAME", "in-cluster")
    except Exception:
        pass

    # Fallback to kubeconfig file
    kubeconfig_path = os.getenv("KUBECONFIG") or str(Path.home() / ".kube" / "config")
    if not os.path.exists(kubeconfig_path):
        return None

    with open(kubeconfig_path, "r") as f:
        kubeconfig = yaml.safe_load(f)
    current_context = kubeconfig.get("current-context")
    for context in kubeconfig.get("contexts", []):
        if context["name"] == current_context:
            return context["context"]["cluster"]
    return None


def print_pod_info(pod_name, pod_idx, is_gpu, metrics=None, queue_name=None):
    """Print pod info with metrics if available"""
    queue_msg = f" | [bold]Queue Name[/bold]: {queue_name}"
    base_msg = f"{BULLET_UNICODE} [reset][bold cyan]{pod_name}[/bold cyan] (idx: {pod_idx})"
    if queue_name:
        base_msg += queue_msg
    console.print(base_msg)
    if metrics:
        console.print(
            f"{DOUBLE_SPACE_UNICODE}[bold]CPU[/bold]: [reset]{metrics['cpu_util']}% "
            f"({metrics['used_cpu']} / {metrics['requested_cpu']}) | "
            f"[bold]Memory[/bold]: {metrics['cpu_memory_usage']} / {metrics['cpu_memory_total']} "
            f"[bold]GB[/bold] ({metrics['cpu_memory_usage_percent']}%)"
        )
        if is_gpu:
            console.print(
                f"{DOUBLE_SPACE_UNICODE}GPU: [reset]{metrics['gpu_util']}% | "
                f"Memory: {metrics['gpu_memory_usage']} / {metrics['gpu_memory_total']} "
                f"GB ({metrics['gpu_memory_usage_percent']}%)"
            )
    else:
        console.print(f"{DOUBLE_SPACE_UNICODE}[yellow]Metrics unavailable[/yellow]")


def _get_logs_from_loki_worker(uri: str, print_pod_name: bool):
    """Worker function for getting logs from Loki - runs in a separate thread."""
    ws = None
    try:
        ws = create_connection(uri)
        message = ws.recv()
        if not message:
            return None
        data = json.loads(message)
        logs = []
        if data.get("streams"):
            for stream in data["streams"]:
                pod_name = f'({stream.get("stream").get("pod")}) ' if print_pod_name else ""
                for value in stream.get("values"):
                    try:
                        log_line = json.loads(value[1])
                        log_name = log_line.get("name")
                        if log_name == "print_redirect":
                            logs.append(f'{pod_name}{log_line.get("message")}')
                        elif log_name != "uvicorn.access":
                            formatted_log = (
                                f"{pod_name}{log_line.get('asctime')} | {log_line.get('levelname')} | "
                                f"{log_line.get('message')}\n"
                            )
                            logs.append(formatted_log)
                    except Exception:
                        logs.append(value[1])
        return logs
    finally:
        if ws:
            try:
                ws.close()
            except Exception:
                pass


def get_logs_from_loki(
    query: str = None,
    uri: str = None,
    print_pod_name: bool = False,
    timeout: float = 5.0,
):
    """Get logs from Loki with fail-fast approach to avoid hanging."""
    try:
        # If URI is provided, use it directly (skip cluster checks)
        if uri:
            return _get_logs_from_loki_worker(uri, print_pod_name)

        import urllib.parse

        # Now safe to proceed with service URL setup
        from kubetorch import globals
        from kubetorch.utils import http_to_ws

        base_url = globals.service_url()
        target_uri = f"{http_to_ws(base_url)}/loki/api/v1/tail?query={urllib.parse.quote_plus(query)}"

        # Use thread timeout for websocket worker since websocket timeouts don't work reliably
        executor = ThreadPoolExecutor(max_workers=1)
        try:
            future = executor.submit(_get_logs_from_loki_worker, target_uri, print_pod_name)
            try:
                result = future.result(timeout=timeout)
                return result
            except TimeoutError:
                logger.debug(f"Loki websocket connection timed out after {timeout}s")
                return None
            except Exception as e:
                logger.debug(f"Error in Loki websocket worker: {e}")
                return None
        finally:
            # Don't wait for stuck threads to complete
            executor.shutdown(wait=False)

    except Exception as e:
        logger.debug(f"Error getting logs from Loki: {e}")
        return None


def stream_logs_websocket(uri, stop_event, service_name, print_pod_name: bool = False):
    """Stream logs using Loki's websocket tail endpoint"""

    console.print(f"\nFollowing logs of [reset]{service_name}\n")

    # Create and run event loop in a separate thread
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        loop.run_until_complete(
            stream_logs_websocket_helper(
                uri=uri,
                stop_event=stop_event,
                stream_type=StreamType.CLI,
                print_pod_name=print_pod_name,
            )
        )
    finally:
        loop.close()
        # Signal the log thread to stop
        stop_event.set()
        # Don't wait for the log thread - it will handle its own cleanup


def get_logs_query(name: str, namespace: str, selected_pod: str, deployment_mode):
    if not selected_pod:
        if deployment_mode in ["knative", "deployment"]:
            # we need to get the pod names first since Loki doesn't have a service_name label
            core_api = client.CoreV1Api()
            pods = validate_pods_exist(name, namespace, core_api)
            pod_names = [pod.metadata.name for pod in pods]
            return f'{{k8s_pod_name=~"{"|".join(pod_names)}",k8s_container_name="kubetorch"}} | json'
        else:
            console.print(f"[red]Logs does not support deployment mode: {deployment_mode}[/red]")
            return None
    else:
        return f'{{k8s_pod_name=~"{selected_pod}",k8s_container_name="kubetorch"}} | json'


def follow_logs_in_cli(
    name: str,
    namespace: str,
    selected_pod: str,
    deployment_mode,
    print_pod_name: bool = False,
):
    """Stream logs when triggerd by the CLI command."""
    from kubetorch.utils import http_to_ws

    stop_event = threading.Event()

    # Set up signal handler to cleanly stop on Ctrl+C
    def signal_handler(signum, frame):
        stop_event.set()
        raise KeyboardInterrupt()

    original_handler = signal.signal(signal.SIGINT, signal_handler)

    # setting up the query
    query = get_logs_query(name, namespace, selected_pod, deployment_mode)
    if not query:
        return
    encoded_query = urllib.parse.quote_plus(query)

    base_url = globals.service_url()
    uri = f"{http_to_ws(base_url)}/loki/api/v1/tail?query={encoded_query}"

    try:
        stream_logs_websocket(
            uri=uri,
            stop_event=stop_event,
            service_name=name,
            print_pod_name=print_pod_name,
        )
    finally:
        # Restore original signal handler
        signal.signal(signal.SIGINT, original_handler)


def is_ingress_vpc_only(annotations: dict):
    # Check for internal LoadBalancer annotations
    internal_checks = [
        annotations.get("service.beta.kubernetes.io/aws-load-balancer-internal") == "true",
        annotations.get("networking.gke.io/load-balancer-type") == "Internal",
        annotations.get("service.beta.kubernetes.io/oci-load-balancer-internal") == "true",
    ]

    vpc_only = any(internal_checks)
    return vpc_only


def load_ingress(namespace: str = globals.config.install_namespace):
    networking_v1_api = client.NetworkingV1Api()
    ingresses = networking_v1_api.list_namespaced_ingress(namespace=namespace)

    for ingress in ingresses.items:
        if ingress.metadata.name == "kubetorch-proxy-ingress":
            return ingress


def get_ingress_host(ingress):
    """Get the configured host from the kubetorch ingress."""
    try:
        return ingress.spec.rules[0].host
    except Exception:
        return None


def list_all_queues():
    try:
        custom_api = client.CustomObjectsApi()
        queues = custom_api.list_cluster_custom_object(
            group="scheduling.run.ai",
            version="v2",
            plural="queues",
        )["items"]

        if not queues:
            console.print("[yellow]No queues found in the cluster[/yellow]")
            return

        # Insert "default" queue if missing
        if not any(q["metadata"]["name"] == "default" for q in queues):
            default_children = [
                q["metadata"]["name"] for q in queues if q.get("spec", {}).get("parentQueue") == "default"
            ]
            queues.insert(
                0,
                {
                    "metadata": {"name": "default"},
                    "spec": {
                        "parentQueue": "-",
                        "children": default_children,
                        "resources": {
                            "cpu": {"quota": "-", "overQuotaWeight": "-"},
                            "gpu": {"quota": "-", "overQuotaWeight": "-"},
                            "memory": {"quota": "-", "overQuotaWeight": "-"},
                        },
                        "priority": "-",
                    },
                },
            )

        queue_table = Table(title="Available Queues", header_style=Style(bold=True))
        queue_table.add_column("QUEUE NAME", style="cyan")
        queue_table.add_column("PRIORITY", style="magenta")
        queue_table.add_column("PARENT", style="green")
        queue_table.add_column("CHILDREN", style="yellow")
        queue_table.add_column("CPU QUOTA", style="white")
        queue_table.add_column("GPU QUOTA", style="white")
        queue_table.add_column("MEMORY QUOTA", style="white")
        queue_table.add_column("OVERQUOTA WEIGHT", style="blue")

        for q in queues:
            spec = q.get("spec", {})
            resources = spec.get("resources", {})
            cpu = resources.get("cpu", {})
            gpu = resources.get("gpu", {})
            memory = resources.get("memory", {})

            queue_table.add_row(
                q["metadata"]["name"],
                str(spec.get("priority", "-")),
                spec.get("parentQueue", "-"),
                ", ".join(spec.get("children", [])) or "-",
                str(cpu.get("quota", "-")),
                str(gpu.get("quota", "-")),
                str(memory.get("quota", "-")),
                str(cpu.get("overQuotaWeight", "-")),  # use CPU's overQuotaWeight as example
            )

        console.print(queue_table)
        return

    except client.exceptions.ApiException as e:
        console.print(f"[red]Failed to list queues: {e}[/red]")
        raise typer.Exit(1)


def detect_deployment_mode(name: str, namespace: str, custom_api, apps_v1_api):
    """Detect if a service is deployed as Knative, Deployment, or RayCluster."""
    # First try Deployment
    try:
        apps_v1_api.read_namespaced_deployment(name=name, namespace=namespace)
        return "deployment"
    except ApiException:
        pass

    # Then try Knative
    try:
        custom_api.get_namespaced_custom_object(
            group="serving.knative.dev",
            version="v1",
            namespace=namespace,
            plural="services",
            name=name,
        )
        return "knative"
    except ApiException:
        pass

    # Then try RayCluster
    try:
        custom_api.get_namespaced_custom_object(
            group="ray.io",
            version="v1",
            namespace=namespace,
            plural="rayclusters",
            name=name,
        )
        return "raycluster"
    except ApiException:
        pass

    return None


def validate_provided_pod(service_name, provided_pod, service_pods):
    if provided_pod is None:
        return provided_pod

    if provided_pod.isnumeric():
        pod = int(provided_pod)
        if pod < 0 or pod >= len(service_pods):
            console.print(f"[red]Pod index {pod} is out of range[/red]")
            raise typer.Exit(1)
        pod_name = service_pods[pod].metadata.name

    # case when the user provides pod name
    else:
        pod_names = [pod.metadata.name for pod in service_pods]
        if provided_pod not in pod_names:
            console.print(f"[red]{service_name} does not have an associated pod called {provided_pod}[/red]")
            raise typer.Exit(1)
        else:
            pod_name = provided_pod

    return pod_name


def load_kubetorch_volumes_for_service(namespace, service_name, core_v1) -> List[str]:
    """Extract volume information from service definition"""
    volumes = []

    try:
        pods = core_v1.list_namespaced_pod(
            namespace=namespace,
            label_selector=f"kubetorch.com/service={service_name}",
        )
        if pods.items:
            pod = pods.items[0]
            for v in pod.spec.volumes or []:
                if v.persistent_volume_claim:
                    volumes.append(v.name)
        return volumes

    except Exception as e:
        logger.warning(f"Failed to extract volumes from service: {e}")

    return volumes


def create_table_for_output(columns: List[set], no_wrap_columns_names: list = None, header_style: dict = None):
    table = Table(box=box.SQUARE, header_style=Style(**header_style))
    for name, style in columns:
        if name in no_wrap_columns_names:
            # always make service name fully visible
            table.add_column(name, style=style, no_wrap=True)
        else:
            table.add_column(name, style=style)

    return table



from enum import Enum
from typing import Any, Dict, List, Union

# Internal class to represent the image construction process
class ImageSetupStepType(Enum):
    """Enum for valid Image setup step types"""

    CMD_RUN = "cmd_run"
    RSYNC = "rsync"
    PIP_INSTALL = "pip_install"
    SYNC_PACKAGE = "sync_package"
    SET_ENV_VARS = "set_env_vars"


class ImageSetupStep:
    def __init__(
        self,
        step_type: ImageSetupStepType,
        **kwargs: Dict[str, Any],
    ):
        """
        A component of the Kubetorch Image, consisting of the step type (e.g. packages, set_env_vars),
        along with arguments to provide to the function corresponding to the step type.

        Args:
            step_type (ImageSetupStepType): Type of setup step used to provide the Image.
            kwargs (Dict[str, Any]): Please refer to the corresponding functions in ``Image`` to determine
                the correct keyword arguments to provide.
        """
        self.step_type = step_type
        self.kwargs = kwargs


class Image:
    def __init__(
        self,
        name: str = None,
        image_id: str = None,
        python_path: str = None,
        install_cmd: str = None,
    ):
        """
        Kubetorch Image object, specifying cluster setup properties and steps.

        Args:
            name (str, optional): Name to assign the Kubetorch image.
            image_id (str, optional): Machine image to use, if any. (Default: ``None``)
            python_path (str, optional): Absolute path to the Python executable to use for remote server and installs.
                (Default: ``None``)
            install_cmd (str, optional): Custom pip/uv install command to use for package installations.
                If not provided, will be inferred based on python_path and available tools (preferring to use uv).
                Examples: "uv pip install", "python -m pip install", "/path/to/.venv/bin/python -m uv pip install"
                (Default: ``None``)

        Note:
            For convenience, Kubetorch provides ready-to-use base images under ``kt.images``.
            These cover common environments like Python, CUDA, and Ray:

              * ``kt.images.Python310()``, ``kt.images.Python311()``, ``kt.images.Python312()``
              * ``kt.images.Debian()``
              * ``kt.images.Ubuntu()``
              * ``kt.images.Ray()`` (defaults to the latest Ray release)

            You can also use flexible factories:

              * ``kt.images.python("3.12")``
              * ``kt.images.pytorch()`` (defaults to "nvcr.io/nvidia/pytorch:23.12-py3")
              * ``kt.images.ray("2.32.0-py311")``

            These base images can be further customized with methods like
            ``.pip_install()``, ``.set_env_vars()``, ``.sync_package()``, etc.

        Example:

            .. code-block:: python

                import kubetorch as kt

                custom_image = (
                    kt.Image(name="base_image")
                    .pip_install(["numpy", "pandas"])
                    .set_env_vars({"OMP_NUM_THREADS": 1})
                )
                debian_image = (
                    kt.images.Debian()
                    .pip_install(["numpy", "pandas"])
                    .set_env_vars({"OMP_NUM_THREADS": 1})
                )
        """

        self.name = name
        self.image_id = image_id
        self.python_path = python_path
        self.install_cmd = install_cmd

        self.setup_steps = []
        self.docker_secret = None

    @staticmethod
    def _setup_step_config(step: ImageSetupStep):
        """Get ImageSetupStep config"""
        config = {
            "step_type": step.step_type.value,
            "kwargs": step.kwargs,
        }
        return config

    @staticmethod
    def _setup_step_from_config(step: Dict):
        """Convert setup step config (dict) to ImageSetupStep object"""
        step_type = step["step_type"]
        kwargs = step["kwargs"]
        return ImageSetupStep(
            step_type=ImageSetupStepType(step_type),
            **kwargs,
        )

    def from_docker(self, image_id: str):
        """Set up and use an existing Docker image.

        Args:
            image_id (str): Docker image in the following format ``"<registry>/<image>:<tag>"``
        """
        if self.image_id:
            raise ValueError("Setting both a machine image and docker image is not yet supported.")
        self.image_id = image_id
        return self

    ########################################################
    # Steps to build the image
    ########################################################

    def pip_install(
        self,
        reqs: List[Union["Package", str]],
        force: bool = False,
    ):
        """Pip install the given packages.

        Args:
            reqs (List[Package or str]): List of packages to pip install on cluster and env.
                Each string is passed directly to the pip/uv command, allowing full control
                over pip arguments. Examples:
                - Simple package: ``"numpy"``
                - Version constraint: ``"pandas>=1.2.0"``
                - With pip flags: ``"--pre torch==2.0.0rc1"``
                - Multiple flags: ``"--index-url https://pypi.org/simple torch"``
            force (bool, optional): Whether to force re-install a package, if it already exists on the compute. (Default: ``False``)

        Example:
            .. code-block:: python

                import kubetorch as kt

                image = (
                    kt.images.Debian()
                    .pip_install([
                        "numpy>=1.20",
                        "pandas",
                        "--pre torchmonarch==0.1.0rc7",  # Install pre-release
                        "--index-url https://test.pypi.org/simple/ mypackage"
                    ])
                )
        """

        self.setup_steps.append(
            ImageSetupStep(
                step_type=ImageSetupStepType.PIP_INSTALL,
                reqs=reqs,
                force=force,
            )
        )
        return self

    def set_env_vars(self, env_vars: Dict):
        """Set environment variables with support for variable expansion.

        Environment variables can reference other variables using shell-style expansion:
        - ``$VAR`` or ``${VAR}`` syntax to reference existing variables
        - Variables are expanded when the container starts
        - Variables are expanded in the order they are defined

        Args:
            env_vars (Dict): Dict of environment variables and values to set.
                Values can include references to other environment variables.

        Example:
            .. code-block:: python

                import kubetorch as kt

                image = (
                    kt.images.Debian()
                    .set_env_vars({
                        "BASE_PATH": "/usr/local",
                        "BIN_PATH": "$BASE_PATH/bin",  # Expands to /usr/local/bin
                        "PATH": "$BIN_PATH:$PATH",      # Prepends to existing PATH
                        "LD_LIBRARY_PATH": "/opt/lib:${LD_LIBRARY_PATH}",  # Appends to existing
                        "CUSTOM": "${HOME}/data",       # Uses HOME from container
                    })
                )

        Note:
            - Variables are expanded using Python's ``os.path.expandvars()``
            - Undefined variables remain as literal strings (e.g., ``$UNDEFINED`` stays as ``$UNDEFINED``)
            - To include a literal ``$``, escape it with backslash: ``\\$``
        """
        # TODO - support .env files
        self.setup_steps.append(
            ImageSetupStep(
                step_type=ImageSetupStepType.SET_ENV_VARS,
                env_vars=env_vars,
            )
        )
        return self

    def sync_package(
        self,
        package: str,
        force: bool = False,
    ):
        """Sync local package over and add to path.

        Args:
            package (Package or str): Package to sync. Either the name of a local editably installed package, or
                the path to the folder to sync over.
            force (bool, optional): Whether to re-sync the package over, if already previously synced over. (Default: ``False``)
        """
        self.setup_steps.append(
            ImageSetupStep(
                step_type=ImageSetupStepType.SYNC_PACKAGE,
                package=package,
                force=force,
            )
        )
        return self

    def run_bash(
        self,
        command: str,
        force: bool = False,
    ):
        """Run bash commands during image setup.

        Executes shell commands during container initialization. Commands run in the
        order they are defined and can be used to install software, configure the
        environment, or start background services.

        Args:
            command (str): Shell command(s) to run on the cluster. Supports:
                - Single commands: ``"apt-get update"``
                - Chained commands: ``"apt-get update && apt-get install -y curl"``
                - Background processes: ``"jupyter notebook --no-browser &"``
            force (bool): Whether to rerun the command on the cluster, if previously run in image setup already. (Default: ``False``)

        Example:
            .. code-block:: python

                import kubetorch as kt

                image = (
                    kt.images.Debian()
                    .run_bash("apt-get update && apt-get install -y vim")
                    .run_bash("pip install jupyter")
                    .run_bash("jupyter notebook --no-browser --port=8888 &")  # Runs in background
                )

        Note:
            - Commands ending with ``&`` run in the background and won't block image setup
            - Background processes continue running after setup completes
            - The setup waits 0.5s to catch immediate failures in background processes
            - Use ``&&`` to chain commands that depend on each other
            - Use ``;`` to run commands sequentially regardless of success/failure
        """
        self.setup_steps.append(
            ImageSetupStep(
                step_type=ImageSetupStepType.CMD_RUN,
                command=command,
                force=force,
            )
        )
        return self

    def rsync(
        self,
        source: str,
        dest: str = None,
        contents: bool = False,
        filter_options: str = None,
        force: bool = False,
    ):
        """Sync files or directories from local machine to the remote container.

        This method efficiently transfers files to remote containers using rsync,
        which only copies changed files and supports compression. Files are first
        uploaded to a jump pod and then distributed to worker pods on startup.

        Args:
            source (str): Path to the local file or directory to sync. Supports:
                - Absolute paths: ``/path/to/file``
                - Relative paths: ``./data/file.txt``
                - Home directory paths: ``~/documents/data.csv``
            dest (str, optional): Target path on the remote container. Supports:
                - Absolute paths: ``/data/config.yaml`` (places file at exact location)
                - Relative paths: ``configs/settings.json`` (relative to working directory)
                - Tilde paths: ``~/results/output.txt`` (relative to working directory, ~ is stripped)
                - None: Uses the basename of source in working directory
            contents (bool, optional): For directories only - whether to copy the contents
                or the directory itself.
                If ``True`` the contents of the source directory are copied to the destination,
                and the source directory itself is not created at the destination.
                If ``False`` the source directory along with its contents are copied to the
                destination, creating an additional directory layer at the destination.
                (Default: ``False``)
            filter_options (str, optional): Additional rsync filter options. These are added
                to (not replacing) the default filters. By default, rsync excludes:

                - Files from ``.gitignore`` (if present)
                - Files from ``.ktignore`` (if present)
                - Common Python artifacts: ``*.pyc``, ``__pycache__``
                - Virtual environments: ``.venv``
                - Git metadata: ``.git``

                Your filter_options are appended after these defaults. Examples:

                - Exclude more patterns: ``"--exclude='*.log' --exclude='temp/'"``
                - Include specific files: ``"--include='important.log' --exclude='*.log'"``
                - Override all defaults: Set ``KT_RSYNC_FILTERS`` environment variable

                (Default: ``None``)
            force (bool, optional): When ``True``, forces rsync to transfer all files
                regardless of modification times by using ``--ignore-times`` flag. This ensures
                all files are copied even if timestamps suggest they haven't changed.
                Useful when timestamp-based change detection is unreliable.
                Note: Files are always synced when deploying with ``.to()``, this flag just
                affects how rsync determines which files need updating.
                (Default: ``False``)

        Returns:
            Image: Returns self for method chaining.

        Examples:
            .. code-block:: python

                import kubetorch as kt

                # Basic file sync
                image = (
                    kt.images.Debian()
                    .rsync("./config.yaml", "app/config.yaml")
                )

                # Sync to absolute path
                image = (
                    kt.images.Python312()
                    .rsync("./model_weights.pth", "/models/weights.pth")
                )

                # No destination specified - uses basename
                image = (
                    kt.images.Ubuntu()
                    .rsync("/local/data/dataset.csv")  # Goes to ./dataset.csv
                )

                # Directory sync - copy directory itself
                image = (
                    kt.images.Debian()
                    .rsync("./src", "app")  # Creates app/src/
                )

                # Directory sync - copy contents only
                image = (
                    kt.images.Debian()
                    .rsync("./src", "app", contents=True)  # Contents go directly into app/
                )

                # Multiple rsync operations with filtering
                image = (
                    kt.images.Python312()
                    .rsync("./data", "/data", filter_options="--exclude='*.tmp'")
                    .rsync("./configs", "~/configs")
                    .rsync("./scripts")
                    .pip_install(["numpy", "pandas"])
                )

                # Force re-sync for development
                image = (
                    kt.images.Debian()
                    .rsync("./rapidly_changing_code", "app", force=True)
                )

        Note:
            - Absolute destination paths (starting with ``/``) place files at exact locations
            - Relative paths and ``~/`` paths are relative to the container's working directory
            - The ``contents`` parameter only affects directory sources, not files
            - Default exclusions include ``.gitignore`` patterns, ``__pycache__``, ``.venv``, and ``.git``
            - User-provided ``filter_options`` are added to (not replacing) the default filters
            - To completely override filters, set the ``KT_RSYNC_FILTERS`` environment variable
            - Use ``force=True`` to bypass timestamp checks and transfer all files
        """

        self.setup_steps.append(
            ImageSetupStep(
                step_type=ImageSetupStepType.RSYNC,
                source=source,
                dest=dest,
                contents=contents,
                filter_options=filter_options,
                force=force,
            )
        )
        return self



import os
import re
import signal
import sys
import threading
import time
from datetime import datetime, timezone
from typing import Dict

from kubetorch.globals import config
from kubetorch.logger import get_logger

from kubetorch.resources.callables.module import Module
from kubetorch.resources.compute.compute import Compute
from kubetorch.resources.compute.utils import ServiceTimeoutError
from kubetorch.servers.http.utils import is_running_in_kubernetes
from kubetorch.utils import get_kt_install_url

logger = get_logger(__name__)


class App(Module):
    MODULE_TYPE = "app"

    def __init__(
        self,
        compute: Compute,
        cli_command: str,
        pointers: tuple,
        name: str = None,
        run_async: bool = False,
    ):
        """
        Initialize an App object for remote execution.

        .. note::

            To create an App, please use the factory method :func:`app` in conjunction with the `kt run` CLI command.

        Args:
            compute (Compute): Compute
            cli_command (str): CLI command to run on the compute.
            pointers (tuple): A tuple containing references needed to locate the app file, of the format
                (current working directory, path of file relative to cwd, None)
            name (str, optional): Name to assign the app. If not provided, will be based on the name of the file in
                which the app was defined.
            run_async (bool, optional): Whether to run the app async. (Default: ``False``)
        """
        super().__init__(name=name, pointers=pointers)
        self.cli_command = cli_command
        self.pointers = pointers
        self.name = name or self.module_name
        self._compute = compute
        self._run_async = run_async
        self._remote_pointers = None

        self._http_client = None

    @property
    def module_name(self):
        return os.path.splitext(self.pointers[1])[0]

    def from_name(self):
        raise ValueError("Reloading app is not supported.")

    def setup_signal_handlers(self):
        signal.signal(signal.SIGINT, self.handle_termination_signal)
        signal.signal(signal.SIGTERM, self.handle_termination_signal)

    def handle_termination_signal(self, signum, frame):
        red = "\u001b[31m"
        reset = "\u001b[0m"

        logger.info(f"{red}Received {signal.Signals(signum).name}. Exiting parent process.{reset}")
        self._print_kt_cmds()
        sys.exit(0)

    def deploy(self):
        """
        Deploy the app to the compute specified by the app arguments.
        """
        self.compute.service_name = self.service_name

        install_url, use_editable = get_kt_install_url(self.compute.freeze)
        if not self.compute.freeze:
            deployment_timestamp = datetime.now(timezone.utc).isoformat()
            self._rsync_repo_and_image_patches(install_url, use_editable, init_args={})
        else:
            deployment_timestamp = None

        self.setup_signal_handlers()

        stream_logs = not self._run_async
        self._launch_service(install_url, use_editable, deployment_timestamp, stream_logs)

    def _get_service_dockerfile(self, metadata_env_vars):
        image_instructions = super()._get_service_dockerfile(metadata_env_vars)

        remote_script = os.path.join(self.remote_pointers[0], self.remote_pointers[1])
        local_script = r"\b" + re.escape(self.remote_pointers[1]) + r"\b"
        remote_cmd = re.sub(local_script, remote_script, self.cli_command)

        image_instructions += f"CMD {remote_cmd}\n"
        return image_instructions

    def _launch_service(
        self,
        install_url,
        use_editable,
        deployment_timestamp,
        stream_logs,
    ):
        trigger_reload = self.compute.is_up()
        if self._run_async:
            thread = threading.Thread(
                target=super()._launch_service,
                args=(
                    install_url,
                    use_editable,
                    {},
                    deployment_timestamp,
                    stream_logs,
                    config.log_verbosity,
                    False,
                ),
            )
            thread.start()

            if trigger_reload:
                self._update_service(stream_logs, deployment_timestamp)
                time.sleep(1)
            else:
                # wait for pods to be ready before exiting out
                start_time = time.time()
                while not self.compute.is_up() and time.time() - start_time < 60:
                    time.sleep(5)

                if not self.compute.is_up():
                    raise ServiceTimeoutError(f"Service {self.service_name} is not up after 60 seconds.")
        else:
            super()._launch_service(
                install_url,
                use_editable,
                init_args={},
                deployment_timestamp=deployment_timestamp,
                stream_logs=stream_logs,
                verbosity=config.log_verbosity,
                dryrun=False,
            )

            if trigger_reload:
                self._update_service(stream_logs, deployment_timestamp)

    def _update_service(self, stream_logs, deployment_timestamp):
        client = self._client()

        if self._run_async:
            thread = threading.Thread(
                target=client.call_method,
                args=(
                    self.endpoint(),
                    stream_logs,
                ),
                kwargs={"headers": {"X-Deployed-As-Of": deployment_timestamp}},
            )
            thread.start()
            time.sleep(1)
            sys.exit()
        else:
            client.call_method(
                self.endpoint(),
                stream_logs=stream_logs,
                headers={"X-Deployed-As-Of": deployment_timestamp},
            )

    def _print_kt_cmds(self):
        logger.info(f"To see logs, run: kt logs {self.service_name}.")
        logger.info(f"To teardown service, run: kt teardown {self.service_name}")

    def endpoint(self):
        return f"{self.base_endpoint}/_reload_image"


def app(
    name: str = None,
    port: int = None,
    health_check: str = None,
    **kwargs: Dict,
):
    """
    Builds and deploys an instance of :class:`App`.

    Args:
        name (str, optional): Name to give the remote app. If not provided, will be based off the name of the file in
            which the app was defined.
        port (int, optional): Server port to expose, if the app starts an HTTP server.
        health_check (str, optional): Health check endpoint, if running a server, to check when server is up and ready.
        **kwargs: Compute kwargs, to define the compute on which to run the app on.

    Examples:

    Define the ``kt.app`` object and compute in your Python file:

    .. code-block:: python

        import kubetorch as kt

        # Define the app at the top of the Python file to deploy
        # train.py
        kt.app(name="my-app", image=kt.Image("docker-latest"), cpus="0.01")

        if __name__ == "__main__":
            ...

    Deploy and run the app remotely using the ``kt run`` CLI command:

    .. code-block:: bash

        kt run python train.py --epochs 5
        kt run fastapi run my_app.py --name fastapi-app
    """
    if not os.getenv("KT_RUN") == "1" or is_running_in_kubernetes():
        return None

    if name and os.getenv("KT_RUN_NAME") and not (name == os.getenv("KT_RUN_NAME")):
        raise ValueError(
            f"Name mismatch between kt.App definition ({name}) and kt run command ({os.getenv('KT_RUN_NAME')})."
        )
    name = name or os.getenv("KT_RUN_NAME")
    cli_command = os.getenv("KT_RUN_CMD")  # set in kt run
    run_async = os.getenv("KT_RUN_ASYNC") == 1

    env_vars = kwargs.get("env_vars", {})
    if port:
        env_vars["KT_APP_PORT"] = port
    if health_check:
        env_vars["KT_APP_HEALTHCHECK"] = health_check
    kwargs["env_vars"] = env_vars
    compute = Compute(**kwargs)

    main_file = os.getenv("KT_RUN_FILE") or os.path.abspath(sys.modules["__main__"].__file__)
    relative_path = os.path.relpath(main_file, os.getcwd())
    pointers = [os.getcwd(), relative_path, None]
    relative_cli_command = re.sub(main_file, relative_path, cli_command)

    kt_app = App(
        compute=compute,
        cli_command=relative_cli_command,
        pointers=pointers,
        name=name,
        run_async=run_async,
    )
    return kt_app



from .. import Secret


class LambdaSecret(Secret):
    """
    .. note::
            To create a LambdaSecret, please use the factory method :func:`secret` with ``provider="lambda"``.
    """

    # values format: {"api_key": api_key}
    _DEFAULT_PATH = "~/.lambda_cloud"
    _DEFAULT_FILENAMES = ["lambda_keys"]
    _PROVIDER = "lambda"



# K8s Configuration
KUBECTL_PORT = 6443
KT_LAUNCH_TIMEOUT = 900  # 15 minutes

# Ports
DEFAULT_NGINX_PORT = 8080
LOCAL_NGINX_PORT = 38080
DEFAULT_KT_SERVER_PORT = 32300  # Standard port of Knative services
DEFAULT_DEBUG_PORT = 5678

# Namespaces
KUBETORCH_NAMESPACE = "kubetorch"
RUNHOUSE_NAMESPACE = "runhouse"
DEFAULT_NAMESPACE = "default"

# Images
SERVER_IMAGE_MINIMAL = "ghcr.io/run-house/server:v3"
SERVER_IMAGE_WITH_OTEL = "ghcr.io/run-house/server-otel:v3"

UBUNTU_IMAGE_MINIMAL = "ghcr.io/run-house/ubuntu:v1"
UBUNTU_IMAGE_WITH_OTEL = "ghcr.io/run-house/ubuntu:v1"

DEFAULT_PROXY_IMAGE = "ghcr.io/run-house/proxy:v2"
KUBETORCH_IMAGE_TRAPDOOR = "kubetorch"

# Service Accounts
DEFAULT_SERVICE_ACCOUNT_NAME = "kubetorch-service-account"

# Annotations
INACTIVITY_TTL_ANNOTATION = "kubetorch.com/inactivity-ttl"
KUBECONFIG_PATH_ANNOTATION = "kubetorch.com/kubeconfig-path"

# Labels
KT_SERVICE_LABEL = "kubetorch.com/service"
KT_VERSION_LABEL = "kubetorch.com/version"
KT_MODULE_LABEL = "kubetorch.com/module"
KT_USER_IDENTIFIER_LABEL = "kubetorch.com/user-identifier"
KT_USERNAME_LABEL = "kubetorch.com/username"
KT_POD_TYPE_LABEL = "kubetorch.com/pod-type"
KT_TEMPLATE_LABEL = "kubetorch.com/template"
KT_SECRET_NAME_LABEL = "kubetorch.com/secret-name"

# Templates
TTL_CONTROLLER_CONFIGMAP_NAME = "kubetorch-ttl-controller-config"
KNATIVE_SERVICE_TEMPLATE_FILE = "knative_service_template.yaml"
POD_TEMPLATE_FILE = "pod_template.yaml"
KT_SETUP_TEMPLATE_FILE = "kt_setup_template.sh.j2"
DEPLOYMENT_TEMPLATE_FILE = "deployment_template.yaml"
DEPLOYMENT_SERVICE_TEMPLATE_FILE = "service_template.yaml"
RAYCLUSTER_TEMPLATE_FILE = "raycluster_template.yaml"
RAYCLUSTER_SERVICE_TEMPLATE_FILE = "raycluster_service_template.yaml"

# Loki
LOKI_GATEWAY_SERVICE_NAME = "loki-gateway"

# Prometheus
PROMETHEUS_SERVICE_NAME = "kubetorch-metrics"

# Grafana
GRAFANA_HEALTH_ENDPOINT = "/api/health"
PROMETHEUS_HEALTH_ENDPOINT = "/metrics"

# KAI
KAI_SCHEDULER_NAME = "kai-scheduler"
KAI_SCHEDULER_LABEL = "kai.scheduler/queue"

# HTTP Client
KT_TERMINATION_REASONS = ["OOMKilled", "Evicted", "Not Found"]

# NGINX
NGINX_GATEWAY_PROXY = "kubetorch-proxy"
DEFAULT_NGINX_HEALTH_ENDPOINT = "/health"

# Rsync
RSYNC_LOCAL_PORT = 3873
REMOTE_RSYNC_PORT = 873
RSYNC_SERVICE_NAME = "kubetorch-rsync"



from .. import Secret


class AzureSecret(Secret):
    """
    .. note::
            To create an AzureSecret, please use the factory method :func:`secret` with ``provider="azure"``.
    """

    # values format: {"subscription_id": subscription_id}
    _PROVIDER = "azure"
    _DEFAULT_PATH = "~/.azure"
    _DEFAULT_FILENAMES = ["clouds.config"]
    _DEFAULT_ENV_VARS = {"subscription_id": "AZURE_SUBSCRIPTION_ID"}



from typing import Union

from .anthropic_secret import AnthropicSecret
from .aws_secret import AWSSecret
from .azure_secret import AzureSecret
from .cohere_secret import CohereSecret
from .gcp_secret import GCPSecret
from .github_secret import GitHubSecret
from .huggingface_secret import HuggingFaceSecret
from .kubeconfig_secret import KubeConfigSecret
from .lambda_secret import LambdaSecret
from .langchain_secret import LangChainSecret
from .openai_secret import OpenAISecret
from .pinecone_secret import PineconeSecret
from .ssh_secret import SSHSecret
from .wandb_secret import WandBSecret

_str_to_provider_class = {
    # File and/or Env secrets
    "aws": AWSSecret,
    "azure": AzureSecret,
    "gcp": GCPSecret,
    "github": GitHubSecret,
    "huggingface": HuggingFaceSecret,
    "kubernetes": KubeConfigSecret,
    "lambda": LambdaSecret,
    # SSH secrets
    "ssh": SSHSecret,
    # API key secrets
    "anthropic": AnthropicSecret,
    "cohere": CohereSecret,
    "langchain": LangChainSecret,
    "openai": OpenAISecret,
    "pinecone": PineconeSecret,
    "wandb": WandBSecret,
}

_path_to_provider_class = {
    "~/.aws": AWSSecret,
    "~/.azure": AzureSecret,
    "~/.config/gcloud": GCPSecret,
    "~/.config/gh": GitHubSecret,
    "~/.cache/huggingface": HuggingFaceSecret,
    "~/.kube": KubeConfigSecret,
    "~/.lambda_cloud": LambdaSecret,
    "~/.ssh": SSHSecret,
}

_secret_to_env_vars = {
    "aws": {"access_key": "AWS_ACCESS_KEY_ID", "secret_key": "AWS_SECRET_ACCESS_KEY"},
    "azure": {"subscription_id": "AZURE_SUBSCRIPTION_ID"},
    "gcp": {"client_id": "CLIENT_ID", "client_secret": "CLIENT_SECRET"},
    "github": {},
    "huggingface": ["HF_TOKEN"],
    "anthropic": {"api_key": "ANTHROPIC_API_KEY"},
    "cohere": {"api_key": "COHERE_API_KEY"},
    "langchain": {"api_key": "LANGCHAIN_API_KEY"},
    "openai": {"api_key": "OPENAI_API_KEY"},
    "pinecone": {"api_key": "PINECONE_API_KEY"},
    "wandb": {"api_key": "WANDB_API_KEY"},
}


def _check_if_provider_secret(provider_str: str = None, provider_env_vars: dict = None):
    import os

    # user passed provider name or path to secrets values
    if provider_str:
        if provider_str in _str_to_provider_class.keys():
            return _str_to_provider_class[provider_str]

        full_default_paths_to_provider = {
            os.path.abspath(path): secret for path, secret in _path_to_provider_class.items()
        }
        provided_full_path = os.path.abspath(provider_str)
        return full_default_paths_to_provider.get(provided_full_path, None)

    # user passed secrets keys to env vars mapping
    elif provider_env_vars:
        for provider_name, default_provider_env_vars in _secret_to_env_vars.items():
            if provider_env_vars == default_provider_env_vars:
                return _str_to_provider_class[provider_name]
        return None


def _get_provider_class(provider_info: Union[str, dict]):
    provider_secret = (
        _check_if_provider_secret(provider_str=provider_info)
        if isinstance(provider_info, str)
        else _check_if_provider_secret(provider_env_vars=provider_info)
    )
    return provider_secret



import subprocess
import textwrap
import uuid

from functools import cached_property
from typing import Dict

from kubernetes import client
from kubernetes.client import ApiException, V1PersistentVolumeClaim

from kubetorch.constants import DEFAULT_VOLUME_ACCESS_MODE, KT_MOUNT_FOLDER
from kubetorch.globals import config
from kubetorch.logger import get_logger
from kubetorch.utils import load_kubeconfig

logger = get_logger(__name__)


class Volume:
    """
    Manages persistent storage for Kubetorch services and deployments.
    """

    def __init__(
        self,
        name: str,
        size: str,
        storage_class: str = None,
        mount_path: str = None,
        access_mode: str = None,
        namespace: str = None,
        core_v1: client.CoreV1Api = None,
    ):
        """
        Kubetorch Volume object, specifying persistent storage properties.

        Args:
            name (str): Name of the volume.
            size (str): Size of the volume.
            storage_class (str, optional): Storage class to use for the volume.
            mount_path (str, optional): Mount path for the volume.
            access_mode (str, optional): Access mode for the volume.
            namespace (str, optional): Namespace for the volume.

        Example:

        .. code-block:: python

            import kubetorch as kt

            kt.Volume(name="my-data", size="5Gi"),  # Standard volume (ReadWriteOnce)

            # Shared volume (ReadWriteMany, requires JuiceFS or similar)
            kt.Volume(name="shared-data", size="10Gi", storage_class="juicefs-sc-shared", access_mode="ReadWriteMany")

            # uv cache
            compute = kt.Compute(
                cpus=".01",
                env_vars={
                    "UV_CACHE_DIR": "/ktfs/kt-global-cache/uv_cache",
                    "HF_HOME": "/ktfs/kt-global-cache/hf_cache",
                },
                volumes=[kt.Volume("kt-global-cache", size="10Gi")],
            )

        """
        self._storage_class = storage_class
        if core_v1 is None:
            load_kubeconfig()

        self.size = size
        self.access_mode = access_mode or DEFAULT_VOLUME_ACCESS_MODE
        self.mount_path = mount_path or f"/{KT_MOUNT_FOLDER}/{name}"

        self.name = name
        self.namespace = namespace
        self.core_v1 = core_v1 or client.CoreV1Api()

    @property
    def pvc_name(self) -> str:
        return self.name

    @cached_property
    def storage_class(self) -> str:
        """Get storage class - either specified or cluster default"""
        if self._storage_class:
            return self._storage_class

        try:
            storage_v1 = client.StorageV1Api()
            storage_classes = storage_v1.list_storage_class().items

            # If RWX is requested, prefer RWX-capable classes
            if self.access_mode == "ReadWriteMany":
                for sc in storage_classes:
                    provisioner = getattr(sc, "provisioner", "")
                    if provisioner in {
                        "csi.juicefs.com",
                        "nfs.csi.k8s.io",
                        "cephfs.csi.ceph.com",
                    }:
                        return sc.metadata.name
                raise ValueError("No RWX-capable storage class found")

            # Otherwise, pick the default StorageClass
            for sc in storage_classes:
                annotations = sc.metadata.annotations or {}
                if annotations.get("storageclass.kubernetes.io/is-default-class") == "true":
                    logger.info(f"Using default storage class: {sc.metadata.name}")
                    return sc.metadata.name

            # No default found, fall back to first available
            available_classes = [sc.metadata.name for sc in storage_classes]
            first_sc = available_classes[0]
            if len(available_classes) == 1:
                logger.info(f"No default storage class found, using only available one: {first_sc}")
            else:
                logger.warning(
                    f"No default storage class found, using first available: {first_sc}. "
                    f"Available: {available_classes}. Consider setting a default or specifying storage_class parameter."
                )
            return first_sc

        except Exception as e:
            logger.error(f"Failed to get storage classes: {e}")
            raise

    @classmethod
    def from_name(
        cls,
        name: str,
        create_if_missing: bool = False,
        namespace: str = None,
        core_v1: client.CoreV1Api = None,
    ) -> "Volume":
        """Get existing volume or optionally create it"""
        if core_v1 is None:
            load_kubeconfig()
            core_v1 = client.CoreV1Api()

        namespace = namespace or config.namespace
        pvc_name = name

        try:
            pvc = core_v1.read_namespaced_persistent_volume_claim(pvc_name, namespace)

            storage_class = pvc.spec.storage_class_name
            size = pvc.spec.resources.requests.get("storage")
            access_mode = pvc.spec.access_modes[0] if pvc.spec.access_modes else DEFAULT_VOLUME_ACCESS_MODE

            annotations = pvc.metadata.annotations or {}
            mount_path = annotations.get("kubetorch.com/mount-path", f"/{KT_MOUNT_FOLDER}/{name}")

            # Create Volume with actual attributes from PVC
            vol = cls(
                name=name,
                storage_class=storage_class,
                mount_path=mount_path,
                size=size,
                access_mode=access_mode,
                namespace=namespace,
                core_v1=core_v1,
            )

            logger.debug(f"Loaded existing PVC {pvc_name} with storage_class={storage_class}")
            return vol

        except ApiException as e:
            if e.status == 404:
                # PVC doesn't exist
                if create_if_missing:
                    vol = cls(name, namespace=namespace, core_v1=core_v1)
                    vol.create()
                    return vol
                else:
                    raise ValueError(f"Volume '{name}' (PVC: {pvc_name}) does not exist in namespace '{namespace}'")
            else:
                # Some other API error
                raise

    def config(self) -> Dict[str, str]:
        """Get configuration for this volume"""
        return {
            "name": self.name,
            "size": self.size,
            "access_mode": self.access_mode,
            "mount_path": self.mount_path,
            "storage_class": self.storage_class,
            "namespace": self.namespace,
        }

    def pod_template_spec(self) -> dict:
        """Convert to Kubernetes volume spec for pod template"""
        return {
            "name": self.name,
            "persistentVolumeClaim": {"claimName": self.pvc_name},
        }

    def create(self) -> V1PersistentVolumeClaim:
        """Create PVC if it doesn't exist"""
        try:
            try:
                # Check if PVC already exists
                existing_pvc = self.core_v1.read_namespaced_persistent_volume_claim(
                    name=self.pvc_name, namespace=self.namespace
                )
                logger.debug(f"PVC {self.pvc_name} already exists in namespace {self.namespace}")
                return existing_pvc
            except ApiException as e:
                if e.status != 404:
                    # Some other error occurred
                    raise

            logger.info(f"Creating new PVC with name: {self.pvc_name}")

            storage_class_name = self.storage_class

            pvc_spec = client.V1PersistentVolumeClaimSpec(
                access_modes=[self.access_mode],
                resources=client.V1ResourceRequirements(requests={"storage": self.size}),
                storage_class_name=storage_class_name,
            )

            pvc_metadata = client.V1ObjectMeta(
                name=self.pvc_name,
                labels={
                    "app": "kubetorch",
                    "kubetorch.com/volume": self.name,
                },
                annotations={"kubetorch.com/mount-path": self.mount_path},
            )

            pvc = client.V1PersistentVolumeClaim(
                api_version="v1",
                kind="PersistentVolumeClaim",
                metadata=pvc_metadata,
                spec=pvc_spec,
            )

            created_pvc = self.core_v1.create_namespaced_persistent_volume_claim(namespace=self.namespace, body=pvc)

            logger.info(
                f"Successfully created PVC {self.pvc_name} in namespace {self.namespace} with "
                f"storage class {storage_class_name}"
            )
            return created_pvc

        except Exception as e:
            logger.error(f"Failed to create PVC {self.pvc_name}: {e}")
            raise

    def delete(self) -> None:
        """Delete the PVC"""
        try:
            self.core_v1.delete_namespaced_persistent_volume_claim(name=self.pvc_name, namespace=self.namespace)
            logger.debug(f"Successfully deleted PVC {self.pvc_name}")
        except ApiException as e:
            if e.status == 404:
                logger.warning(f"PVC {self.pvc_name} not found")
            else:
                logger.error(f"Failed to delete PVC {self.pvc_name}: {e}")
                raise

    def exists(self) -> bool:
        """Check if the PVC exists"""
        try:
            self.core_v1.read_namespaced_persistent_volume_claim(name=self.pvc_name, namespace=self.namespace)
            return True
        except ApiException as e:
            if e.status == 404:
                return False
            else:
                # Some other API error, re-raise
                raise

    def ssh(self, image: str = "alpine:latest"):
        """
        Launch an interactive debug shell with this volume mounted.

        This method creates a temporary Kubernetes pod that mounts the
        PersistentVolumeClaim (PVC) backing this Volume at the same path
        (`self.mount_path`) used by Kubetorch services.

        Args:
            image (str, optional): Container image to use for the debug pod.
                Must include a shell (e.g., `alpine:3.18`, `ubuntu:22.04`,
                or a custom tools image). Defaults to `alpine:latest`.

        Example:

        .. code-block:: python

            import kubetorch as kt

            vol = kt.Volume.from_name("kt-global-cache")
            vol.ssh()
        """
        pod_name = f"debug-{self.name}-{uuid.uuid4().hex[:6]}"

        cmd = [
            "kubectl",
            "run",
            pod_name,
            "--rm",
            "-it",
            "--namespace",
            self.namespace,
            "--image",
            image,
            "--restart=Never",
            "--overrides",
            textwrap.dedent(
                f"""
            {{
              "apiVersion": "v1",
              "spec": {{
                "containers": [{{
                  "name": "debug",
                  "image": "{image}",
                  "stdin": true,
                  "tty": true,
                  "volumeMounts": [{{
                    "name": "vol",
                    "mountPath": "{self.mount_path}"
                  }}]
                }}],
                "volumes": [{{
                  "name": "vol",
                  "persistentVolumeClaim": {{
                    "claimName": "{self.pvc_name}"
                  }}
                }}]
              }}
            }}
            """
            ).strip(),
        ]

        # Suppress noisy "write on closed stream" error when exiting
        subprocess.run(cmd, stderr=subprocess.DEVNULL)



# Configuration file for the Sphinx documentation builder.

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath("../"))

# -- Project information -----------------------------------------------------

project = "Kubetorch"
copyright = "Runhouse Inc"
author = "the Runhouse team 🏃‍♀️🏠"

# The full version, including alpha/beta/rc tags
import kubetorch

release = kubetorch.__version__


# -- General configuration ---------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "myst_parser",
    "_ext.json_globaltoc",
]

autodoc_typehints_format = "short"
autodoc_default_flags = ["members", "show-inheritance"]
autodoc_member_order = "bysource"

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

markdown_http_base = "/docs/guide"
markdown_anchor_sections = True

if tags.has("json"):
    # Force simpler output format (helps CLI output)
    autodoc_typehints = "signature"  # "description"
    napoleon_use_param = True
    napoleon_use_rtype = True

    html_link_suffix = ""
    json_baseurl = "docs/"

# -- Options for HTML output -------------------------------------------------

if not tags.has("json"):
    html_theme = "sphinx_book_theme"

html_title = "Kubetorch"
html_theme_options = {
    "path_to_docs": "docs/",
    "home_page_in_toc": True,
}

# -- Disable "View Source" links and code display ----------------------------

html_show_sourcelink = False  # hides "View Source" link
html_copy_source = False  # prevents .html files from containing source code



import base64
import importlib
import importlib.util
import inspect
import json
import logging.config
import os
import pickle
import random
import subprocess
import sys
import threading
import time
import traceback
from contextlib import asynccontextmanager
from datetime import datetime, timezone
from pathlib import Path
from typing import Awaitable, Callable, Dict, Optional, Union

try:
    import httpx
except:
    pass

from fastapi import Body, FastAPI, Header, HTTPException, Request

from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse

from pydantic import BaseModel
from starlette.middleware.base import BaseHTTPMiddleware

try:
    from server_metrics import get_inactivity_ttl_annotation, HeartbeatManager, setup_otel_metrics
    from utils import (
        clear_debugging_sessions,
        deep_breakpoint,
        DEFAULT_ALLOWED_SERIALIZATION,
        ensure_structured_logging,
        is_running_in_kubernetes,
        LOG_CONFIG,
        request_id_ctx_var,
        RSYNC_PORT,
        wait_for_app_start,
    )
except ImportError:
    from .server_metrics import get_inactivity_ttl_annotation, HeartbeatManager, setup_otel_metrics
    from .utils import (
        clear_debugging_sessions,
        deep_breakpoint,
        DEFAULT_ALLOWED_SERIALIZATION,
        ensure_structured_logging,
        is_running_in_kubernetes,
        LOG_CONFIG,
        request_id_ctx_var,
        RSYNC_PORT,
        wait_for_app_start,
    )

from starlette.background import BackgroundTask
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.responses import StreamingResponse

logging.config.dictConfig(LOG_CONFIG)

# Set up our structured JSON logging
ensure_structured_logging()

# Create the print logger AFTER ensure_structured_logging so it inherits handlers
print_logger = logging.getLogger("print_redirect")

logger = logging.getLogger(__name__)
# Set log level based on environment variable
# Don't default the log_level
kt_log_level = os.getenv("KT_LOG_LEVEL")
if kt_log_level:
    kt_log_level = kt_log_level.upper()
    logger.setLevel(getattr(logging, kt_log_level, logging.INFO))

_CACHED_CALLABLES = {}
_LAST_DEPLOYED = 0
_CACHED_IMAGE = []
DISTRIBUTED_SUPERVISOR = None
APP_PROCESS = None
_CALLABLE_LOAD_LOCK = threading.Lock()  # Lock for thread-safe callable loading
LOKI_HOST = os.environ.get("LOKI_HOST", "loki-gateway.kubetorch.svc.cluster.local")
LOKI_PORT = int(os.environ.get("LOKI_PORT", 80))  # Default Loki port
KT_OTEL_ENABLED = os.environ.get("KT_OTEL_ENABLED", "False").lower() == "true"

# Global termination event that can be checked by running requests
TERMINATION_EVENT = threading.Event()
# Create a client for FastAPI service

# Set the python breakpoint to kt.deep_breakpoint
os.environ["PYTHONBREAKPOINT"] = "kubetorch.deep_breakpoint"

request_id_ctx_var.set(os.getenv("KT_LAUNCH_ID", "-"))

#####################################
########### Proxy Helpers ###########
#####################################
if os.getenv("KT_CALLABLE_TYPE") == "app" and os.getenv("KT_APP_PORT"):
    port = os.getenv("KT_APP_PORT")
    logger.info(f"Creating /http reverse proxy to: http://localhost:{port}/")
    proxy_client = httpx.AsyncClient(base_url=f"http://localhost:{port}/", timeout=None)
else:
    proxy_client = None


async def _http_reverse_proxy(request: Request):
    """Reverse proxy for /http/* routes to FastAPI service on its port"""
    # Extract the endpoint name from the path
    # request.path_params["path"] will contain everything after /http/
    endpoint_path = request.path_params["path"]

    # Build the URL for the FastAPI service
    url = httpx.URL(path=f"/{endpoint_path}", query=request.url.query.encode("utf-8"))

    # Build the request to forward to FastAPI
    rp_req = proxy_client.build_request(request.method, url, headers=request.headers.raw, content=await request.body())

    # Send the request and get streaming response
    rp_resp = await proxy_client.send(rp_req, stream=True)

    # Return streaming response
    return StreamingResponse(
        rp_resp.aiter_raw(),
        status_code=rp_resp.status_code,
        headers=rp_resp.headers,
        background=BackgroundTask(rp_resp.aclose),
    )


#####################################
########### Cache Helpers ###########
#####################################
def clear_cache():
    global _CACHED_CALLABLES

    logger.debug("Clearing callables cache.")
    _CACHED_CALLABLES.clear()


def cached_image_setup():
    logger.debug("Starting cached image setup.")
    global _CACHED_IMAGE
    global APP_PROCESS

    dockerfile_path = kt_directory() / "image.dockerfile"
    with open(dockerfile_path, "r") as file:
        lines = file.readlines()
    lines = [line.strip() for line in lines]

    # find first line where image differs from cache and update cache
    cache_mismatch_index = -1
    cmd_mismatch = False
    for i, (new_line, cached_line) in enumerate(zip(lines, _CACHED_IMAGE)):
        if new_line.startswith("CMD"):
            cmd_mismatch = True

        if new_line != cached_line or "# override" in new_line or cmd_mismatch:
            cache_mismatch_index = i
            break
    if cache_mismatch_index == -1:
        if len(lines) != len(_CACHED_IMAGE):
            cache_mismatch_index = min(len(lines), len(_CACHED_IMAGE))
        else:
            cache_mismatch_index = len(lines)
    _CACHED_IMAGE = lines

    if cache_mismatch_index == len(lines):
        return

    if not (cache_mismatch_index == len(lines) - 1 and cmd_mismatch):
        logger.info("Running image setup.")
    else:
        logger.debug("Skipping image setup steps, no changes detected.")

    # Grab the current list of installed dependencies with pip freeze to check if anything changes (we need to send a
    # SIGHUP to restart the server if so)
    start_deps = None
    import subprocess

    try:
        res = subprocess.run(
            ["pip", "freeze"],
            capture_output=True,
            text=True,
            check=True,
        )
        start_deps = res.stdout.splitlines()
    except subprocess.CalledProcessError as e:
        logger.error(f"Failed to run pip freeze: {e}")

    # only run image setup steps starting from cache mismatch point
    kt_pip_cmd = None
    for line in lines[cache_mismatch_index:]:
        command = ""
        if line.strip().startswith("#"):
            continue  # ignore comments
        if line.startswith("RUN") or line.startswith("CMD"):
            command = line[len("RUN ") :]

            if command.startswith("$KT_PIP_INSTALL_CMD"):
                kt_pip_cmd = kt_pip_cmd or _get_kt_pip_install_cmd()
                command = command.replace("$KT_PIP_INSTALL_CMD", kt_pip_cmd)
        elif line.startswith("COPY"):
            _, source, dest = line.split()
            # COPY instructions are essentially no-ops since rsync_file_updates()
            # already placed files in their correct locations.
            # But we verify the files exist and log the absolute paths for clarity.

            # Determine the actual absolute destination path
            if dest and dest.startswith("/"):
                # Already absolute
                dest_path = Path(dest)
            elif dest and dest.startswith("~/"):
                # Tilde prefix - strip it and treat as relative to cwd
                dest_path = Path.cwd() / dest[2:]
            else:
                # Relative to working directory (including explicit basenames)
                dest_path = Path.cwd() / dest

            # Verify the destination exists (it should have been rsync'd)
            if dest_path.exists():
                logger.info(f"Copied {source} to {dest_path.absolute()}")
            else:
                raise FileNotFoundError(
                    f"COPY {source} {dest} failed: destination {dest_path.absolute()} does not exist. "
                    f"This likely means the rsync operation failed to sync the files correctly."
                )
        elif line.startswith("ENV"):
            # Need to handle the case where the env var is being set to "" (empty string)
            line_vals = line.split(" ", 2)
            if len(line_vals) < 2:  # ENV line must have at least key
                raise ValueError("ENV line cannot be empty")
            if len(line_vals) == 2:  # ENV line with just key
                key = line_vals[1]
                val = ""
            elif len(line_vals) == 3:  # ENV line with key and value
                key, val = line_vals[1], line_vals[2]

            # Expand environment variables in the value
            # This supports patterns like $VAR, ${VAR}, and $VAR:default_value
            expanded_val = os.path.expandvars(val)

            if key not in [
                "KT_FILE_PATH",
                "KT_MODULE_NAME",
                "KT_CLS_OR_FN_NAME",
                "KT_INIT_ARGS",
                "KT_CALLABLE_TYPE",
                "KT_DISTRIBUTED_CONFIG",
            ]:
                logger.info(f"Setting env var {key}")
            os.environ[key] = expanded_val
            # If the env var is specifically KT_LOG_LEVEL, we need to update the logger level
            if key == "KT_LOG_LEVEL":
                global kt_log_level
                kt_log_level = expanded_val.upper()
                logger.setLevel(kt_log_level)
                logger.info(f"Updated log level to {kt_log_level}")
        elif line.startswith("FROM"):
            continue
        elif line:
            raise ValueError(f"Unrecognized image setup instruction {line}")

        if command:
            is_app_cmd = line.startswith("CMD")
            if is_app_cmd:
                logger.info(f"Running app command: {command}")
            else:
                logger.info(f"Running image setup step: {command}")

            try:
                # Use subprocess.Popen to capture output and redirect through StreamToLogger
                env = os.environ.copy()
                env["PYTHONUNBUFFERED"] = "1"

                if is_app_cmd and os.getenv("KT_CALLABLE_TYPE") == "app":
                    if APP_PROCESS and APP_PROCESS.poll() is None:
                        APP_PROCESS.kill()

                process = subprocess.Popen(
                    command,
                    shell=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    universal_newlines=True,
                    bufsize=1,
                    env=env,
                )

                if is_app_cmd and os.getenv("KT_CALLABLE_TYPE") == "app":
                    APP_PROCESS = process

                # Collect stderr for potential error logging
                import threading

                stderr_lines = []
                stderr_lock = threading.Lock()

                # Stream stdout and stderr in real-time
                # We need to do all this so the stdout and stderr are prints with the correct formatting
                # for our queries. Without it they just flow straight to system stdout and stderr without any

                def stream_output(pipe, log_func, request_id, collect_stderr=False):
                    request_id_ctx_var.set(request_id)
                    for line in iter(pipe.readline, ""):
                        if line:
                            stripped_line = line.rstrip()
                            log_func(stripped_line)

                            # Collect stderr lines for potential error logging
                            if collect_stderr:
                                with stderr_lock:
                                    stderr_lines.append(stripped_line.lstrip("ERROR: "))
                    pipe.close()

                # Start streaming threads
                current_request_id = request_id_ctx_var.get("-")

                stderr_log_func = logger.error if is_app_cmd else logger.debug
                stdout_thread = threading.Thread(
                    target=stream_output,
                    args=(process.stdout, logger.info, current_request_id),
                )
                stderr_thread = threading.Thread(
                    target=stream_output,
                    args=(
                        process.stderr,
                        stderr_log_func,
                        current_request_id,
                        not is_app_cmd,
                    ),
                )

                stdout_thread.daemon = True
                stderr_thread.daemon = True
                stdout_thread.start()
                stderr_thread.start()

                if is_app_cmd and os.getenv("KT_APP_PORT"):
                    # wait for internal app to be healthy/ready if run port is provided
                    try:
                        port = os.getenv("KT_APP_PORT")
                        logger.debug(f"Waiting for internal app on port {port} to start:")
                        wait_for_app_start(
                            port=port,
                            health_check=os.getenv("KT_APP_HEALTHCHECK"),
                            process=process,
                        )
                        logger.info(f"App on port {port} is ready.")
                    except Exception as e:
                        logger.error(f"Caught exception waiting for app to start: {e}")
                else:
                    # Check if this is a background command (ends with &)
                    is_background = command.rstrip().endswith("&")

                    if is_background:
                        # For background processes, give it a moment to start and check for immediate failures
                        import time

                        time.sleep(0.5)  # Brief pause to catch immediate errors

                        # Check if process failed immediately
                        poll_result = process.poll()
                        if poll_result is not None and poll_result != 0:
                            # Process exited with error
                            stdout_thread.join(timeout=1)
                            stderr_thread.join(timeout=1)
                            return_code = poll_result
                        else:
                            # Process is running in background successfully
                            logger.info(f"Background process started successfully (PID: {process.pid})")
                            return_code = 0  # Indicate success for background start
                    else:
                        # Wait for process to complete
                        return_code = process.wait()

                        # Wait for streaming threads to finish
                        stdout_thread.join()
                        stderr_thread.join()

                    if return_code != 0 and not is_app_cmd:
                        with stderr_lock:
                            if stderr_lines:
                                logger.error(f"Failed to run command '{command}' with stderr:")
                                for stderr_line in stderr_lines:
                                    logger.error(stderr_line)
            except subprocess.CalledProcessError as e:
                logger.error(f"Failed to run command '{command}' with error: {e}")
                with stderr_lock:
                    if stderr_lines:
                        logger.error("Stderr:")
                        for stderr_line in stderr_lines:
                            logger.error(stderr_line)
    # Check if any dependencies changed and if so reload them inside the server process
    if start_deps:
        try:
            # Run pip freeze and capture the output
            res = subprocess.run(
                ["pip", "freeze"],
                capture_output=True,
                text=True,
                check=True,
            )
            end_deps = res.stdout.splitlines()
            # We only need to look at the deps which were already installed (i.e. lines in start_deps),
            # new ones can't be "stale" inside the current server process
            # We also only use lines with exact pypi versions (has "=="), no editable
            changed_deps = [line.split("==")[0] for line in start_deps if "==" in line and line not in end_deps]
            imported_changed_deps = [
                dep for dep in changed_deps if dep in sys.modules
            ]  # Only reload deps which are already imported
            if imported_changed_deps:
                logger.debug(f"New dependencies found: {imported_changed_deps}, forcing reload")

                # Don't clear the callable cache here - let load_callable_from_env handle it to preserve __kt_cached_state__
                if DISTRIBUTED_SUPERVISOR:
                    DISTRIBUTED_SUPERVISOR.cleanup()

                # Remove changed modules from sys.modules to override fresh imports
                modules_to_remove = []
                for module_name in sys.modules:
                    for dep in imported_changed_deps:
                        if module_name == dep or module_name.startswith(dep + "."):
                            modules_to_remove.append(module_name)
                            break

                for module_name in modules_to_remove:
                    try:
                        del sys.modules[module_name]
                        logger.debug(f"Removed module {module_name} from sys.modules")
                    except KeyError:
                        pass
        except subprocess.CalledProcessError as e:
            logger.error(f"Failed to run pip freeze: {e}")


def run_image_setup(deployed_time: Optional[float] = None):
    if os.environ["KT_FREEZE"] == "True" or not is_running_in_kubernetes():
        return

    rsync_file_updates()

    dockerfile_path = kt_directory() / "image.dockerfile"
    if not dockerfile_path.exists():
        raise FileNotFoundError(f"No image and metadata configuration found in path: {str(dockerfile_path)}")
    while (
        # May need to give the dockerfile time to rsync over, so wait until the dockerfile timestamp is later than
        # when we started the deployment (recorded in .to and passed here as deployed_time). We also should only
        # wait if _LAST_DEPLOYED is not zero, as the first time the server is deployed the image is written before
        # the server starts so we don't need to wait.
        _LAST_DEPLOYED
        and dockerfile_path.stat().st_mtime < deployed_time
        and datetime.now(timezone.utc).timestamp() - deployed_time < 5
    ):
        time.sleep(0.1)

    cached_image_setup()

    if not os.getenv("KT_CALLABLE_TYPE") == "app":
        logger.debug("Completed cached image setup.")


#####################################
######## Generic Helpers ############
#####################################
class SerializationError(Exception):
    pass


def kt_directory():
    if "KT_DIRECTORY" in os.environ:
        return Path(os.environ["KT_DIRECTORY"]).expanduser()
    else:
        return Path.cwd() / ".kt"


def _get_kt_pip_install_cmd() -> Optional[str]:
    """Get the actual KT_PIP_INSTALL_CMD value for command expansion."""
    kt_pip_cmd = os.getenv("KT_PIP_INSTALL_CMD")
    if not kt_pip_cmd:  # Fallback to reading from file
        try:
            with open(kt_directory() / "kt_pip_install_cmd", "r") as f:
                return f.read().strip()
        except FileNotFoundError:
            return None
    return kt_pip_cmd


def is_running_in_container():
    # Check for .dockerenv file which exists in Docker containers
    return Path("/.dockerenv").exists()


async def run_in_executor_with_context(executor, func, *args):
    """
    Helper to run a function in an executor while preserving the request_id context.

    This wrapper captures the current request_id from the context before running
    the function in a thread pool executor, then sets it in the new thread.
    """
    import asyncio

    # Capture the current request_id before switching threads
    current_request_id = request_id_ctx_var.get("-")

    def wrapper(*args):
        # Set the request_id in the executor thread
        token = None
        if current_request_id != "-":
            token = request_id_ctx_var.set(current_request_id)
        try:
            return func(*args)
        finally:
            # Clean up the context to avoid leaking between requests
            if token is not None:
                request_id_ctx_var.reset(token)

    return await asyncio.get_event_loop().run_in_executor(executor, wrapper, *args)


def should_reload(deployed_as_of: Optional[str] = None) -> bool:
    """
    Determine if the server should reload based on the deployment timestamp.
    If deployed_as_of is provided, it checks against the last deployed time.
    If not provided, it defaults to False.
    """
    if deployed_as_of in [None, "null", "None"]:
        return False

    try:
        deployed_time = datetime.fromisoformat(deployed_as_of).timestamp()
        return deployed_time > _LAST_DEPLOYED
    except ValueError as e:
        logger.error(f"Invalid deployed_as_of format: {deployed_as_of}. Error: {e}")
        return True


def load_callable(
    deployed_as_of: Optional[str] = None,
    distributed_subprocess: bool = False,
    reload_cleanup_fn: [Callable, None] = None,
):
    global _LAST_DEPLOYED

    callable_name = os.environ["KT_CLS_OR_FN_NAME"]

    callable_obj = _CACHED_CALLABLES.get(callable_name, None)
    if callable_obj and not should_reload(deployed_as_of):
        # If the callable is cached and doesn't need reload, return it immediately
        logger.debug("Returning cached callable.")
        return callable_obj

    # Slow path: need to load or reload - use lock for thread safety
    with _CALLABLE_LOAD_LOCK:
        # Double-check within lock (another thread might have loaded it)
        callable_obj = _CACHED_CALLABLES.get(callable_name, None)
        if callable_obj and not should_reload(deployed_as_of):
            logger.debug("Returning cached callable (found after acquiring lock).")
            return callable_obj
        # Proceed with loading/reloading
        return _load_callable_internal(deployed_as_of, distributed_subprocess, reload_cleanup_fn, callable_obj)


def _load_callable_internal(
    deployed_as_of: Optional[str] = None,
    distributed_subprocess: bool = False,
    reload_cleanup_fn: [Callable, None] = None,
    callable_obj=None,
):
    """Internal callable loading logic - should be called within lock for thread safety."""
    global _LAST_DEPLOYED

    callable_name = os.environ["KT_CLS_OR_FN_NAME"]

    if not callable_obj:
        logger.debug("Callable not found in cache, loading from environment.")
    else:
        logger.debug(
            f"Callable found in cache, but reloading because deployed_as_of {deployed_as_of} is newer than last deployed time {_LAST_DEPLOYED}"
        )

    # If not in cache or we have a more recent deployment timestamp, update metadata and reload
    if reload_cleanup_fn and _LAST_DEPLOYED:
        # If a reload cleanup function is provided and we've already deployed at least once, call it before
        # reloading the callable
        reload_cleanup_fn()

    deployed_time = (
        datetime.fromisoformat(deployed_as_of).timestamp() if deployed_as_of else datetime.now(timezone.utc).timestamp()
    )
    if not distributed_subprocess:
        # We don't reload the image in distributed subprocess/es, as we already did it in the
        # main process and we don't want to do it multiple times (in each subprocess).
        if _LAST_DEPLOYED:
            logger.info("Patching image and code updates and reloading callable.")
        else:
            logger.info("Setting up image and loading callable.")
        run_image_setup(deployed_time)

    distributed_config = os.environ["KT_DISTRIBUTED_CONFIG"]
    if distributed_config not in ["null", "None"] and not distributed_subprocess:
        logger.debug(f"Loading distributed supervisor: {distributed_config}")
        callable_obj = load_distributed_supervisor(deployed_as_of=deployed_as_of)
        logger.debug("Distributed supervisor loaded successfully.")
    else:
        logger.debug(f"Loading callable from environment: {callable_name}")
        callable_obj = load_callable_from_env()
        logger.debug("Callable loaded successfully.")

    _LAST_DEPLOYED = deployed_time
    _CACHED_CALLABLES[callable_name] = callable_obj

    return callable_obj


def load_distributed_supervisor(deployed_as_of: Optional[str] = None):
    global DISTRIBUTED_SUPERVISOR

    if os.environ["KT_FILE_PATH"] not in sys.path:
        sys.path.insert(0, os.environ["KT_FILE_PATH"])

    distributed_config = os.environ["KT_DISTRIBUTED_CONFIG"]

    # If this is the main process of a distributed call, we don't load the callable directly,
    # we create a new supervisor if it doesn't exist or if the config has changed.
    # We don't create a supervisor if this is a distributed subprocess.
    config_hash = hash(str(distributed_config))
    if DISTRIBUTED_SUPERVISOR is None or config_hash != DISTRIBUTED_SUPERVISOR.config_hash:
        from .distributed_utils import distributed_supervisor_factory

        logger.info(f"Loading distributed supervisor with config: {distributed_config}")
        distributed_config = json.loads(distributed_config)
        # If we already have some distributed processes, we need to clean them up before creating a new supervisor.
        if DISTRIBUTED_SUPERVISOR:
            DISTRIBUTED_SUPERVISOR.cleanup()
        DISTRIBUTED_SUPERVISOR = distributed_supervisor_factory(**distributed_config)
        DISTRIBUTED_SUPERVISOR.config_hash = config_hash
    try:
        # If there are any errors during setup, we catch and log them, and then undo the setup
        # so that the distributed supervisor is not left in a broken state (and otherwise can still fail
        # when we call DISTRIBUTED_SUPERVISOR.cleanup() in lifespan).
        DISTRIBUTED_SUPERVISOR.setup(deployed_as_of=deployed_as_of)
    except Exception as e:
        logger.error(f"Failed to set up distributed supervisor with config {distributed_config}: {e}")
        DISTRIBUTED_SUPERVISOR = None
        raise e
    return DISTRIBUTED_SUPERVISOR


def patch_sys_path():
    abs_path = str(Path(os.environ["KT_FILE_PATH"]).expanduser().resolve())
    if os.environ["KT_FILE_PATH"] not in sys.path:
        sys.path.insert(0, abs_path)
        logger.debug(f"Added {abs_path} to sys.path")

    # Maybe needed for subprocesses (e.g. distributed) to find the callable's module
    # Needed for distributed subprocesses to find the file path
    existing_path = os.environ.get("PYTHONPATH", "")
    if os.environ["KT_FILE_PATH"] not in existing_path:
        os.environ["PYTHONPATH"] = f"{abs_path}{os.pathsep}{existing_path}" if existing_path else abs_path
        logger.debug(f"Set PYTHONPATH to {os.environ['PYTHONPATH']}")


def load_callable_from_env():
    """Load and cache callable objects from env, preserving state if __kt_cached_state__ is available."""
    cls_or_fn_name = os.environ["KT_CLS_OR_FN_NAME"]
    module_name = os.environ["KT_MODULE_NAME"]

    # Check if we have an existing cached callable and extract state if available
    cached_state = None
    existing_callable = _CACHED_CALLABLES.get(cls_or_fn_name, None)

    if existing_callable and hasattr(existing_callable, "__kt_cached_state__"):
        try:
            logger.info(f"Extracting cached state from {cls_or_fn_name} via __kt_cached_state__")
            cached_state = existing_callable.__kt_cached_state__()
            if cached_state is not None and not isinstance(cached_state, dict):
                logger.warning(
                    f"__kt_cached_state__ returned non-dict type: {type(cached_state)}. Ignoring cached state."
                )
                cached_state = None
        except Exception as e:
            # This could happen if modules were removed from sys.modules during image setup
            # and the callable's __kt_cached_state__ method depends on them
            logger.warning(
                f"Failed to extract cached state from {cls_or_fn_name} (possibly due to module reloading): {e}. "
                f"Proceeding without cached state."
            )
            cached_state = None

    # Now that we have the state, clean up the old callable to free memory
    if existing_callable:
        logger.debug(f"Deleting existing callable: {cls_or_fn_name}")
        _CACHED_CALLABLES.pop(cls_or_fn_name, None)
        del existing_callable
        # Garbage collect to ensure everything cleaned up (especially GPU memory)
        import gc

        gc.collect()

    patch_sys_path()

    # If we're inside a distributed subprocess or the main process of a non-distributed call,
    # we load and instantiate the callable.
    try:
        # Try regular package import first
        if module_name in sys.modules:
            # Clear any existing debugging sessions when reloading modules
            clear_debugging_sessions()

            # Remove the module and any modules under KT_FILE_PATH from sys.modules
            # This ensures we reload all user updates including cached submodules that were imported by the main module.
            logger.info(f"Removing module {module_name} and submodules from sys.modules")
            kt_file_path = os.environ.get("KT_FILE_PATH")
            kt_file_path_str = str(Path(kt_file_path).expanduser().resolve()) + os.sep

            del sys.modules[module_name]
            for name, mod in list(sys.modules.items()):
                if name == module_name or not hasattr(mod, "__file__") or not mod.__file__:
                    continue
                try:
                    if os.path.abspath(mod.__file__).startswith(kt_file_path_str):
                        del sys.modules[name]
                except (KeyError, OSError, ValueError):
                    pass

            # We make this log to info because some imports are slow and we want the user to know that
            # it's not our fault and not hanging
            logger.info(f"Reimporting module {module_name} and related submodules")
        else:
            logger.debug(f"Importing module {module_name}")

        module = importlib.import_module(module_name)
        logger.debug(f"Module {module_name} loaded")

        # Ensure our structured logging is in place after user module import
        # (in case the user's module configured its own logging)
        ensure_structured_logging()

        callable_obj = getattr(module, cls_or_fn_name)
        logger.debug(f"Callable {cls_or_fn_name} loaded")
    except (ImportError, ValueError) as original_error:
        # Fall back to file-based import if package import fails
        try:
            module = import_from_file(os.environ["KT_FILE_PATH"], module_name)
            # Ensure structured logging after file-based import
            ensure_structured_logging()
            callable_obj = getattr(module, cls_or_fn_name)
        except (ImportError, ValueError):
            # Raise the original error if file import also fails, because the errors which are raised here are
            # more opaque and less useful than the original ImportError or ValueError.
            raise original_error
    except AttributeError as e:
        # If the callable is not found in the module, raise an error
        raise HTTPException(
            status_code=404,
            detail=f"Callable '{cls_or_fn_name}' not found in module '{module_name}'",
        ) from e

    # Unwrap to remove any kt deploy decorators (e.g. @kt.compute)
    if hasattr(callable_obj, "__wrapped__"):
        callable_obj = callable_obj.__wrapped__

    if isinstance(callable_obj, type):
        # Prepare init arguments
        init_kwargs = {}

        # Add user-provided init_args
        if os.environ["KT_INIT_ARGS"] not in ["null", "None"]:
            init_kwargs = json.loads(os.environ["KT_INIT_ARGS"])
            logger.info(f"Setting init_args {init_kwargs}")

        # Add cached state if available
        # Allow user to manually set "kt_cached_state" to override/disable cache
        if cached_state is not None and "kt_cached_state" not in init_kwargs:
            # Check if the class's __init__ accepts kt_cached_state parameter
            sig = inspect.signature(callable_obj.__init__)
            if "kt_cached_state" in sig.parameters:
                logger.info(f"Passing cached state to {cls_or_fn_name}.__init__")
                init_kwargs["kt_cached_state"] = cached_state
            else:
                raise ValueError(
                    f"Class {cls_or_fn_name} has __kt_cached_state__ method but __init__ does not accept "
                    f"'kt_cached_state' parameter. Please add 'kt_cached_state=None' to __init__ signature."
                )

        # Instantiate with combined arguments
        if init_kwargs:
            callable_obj = callable_obj(**init_kwargs)
        else:
            callable_obj = callable_obj()

    return callable_obj


def import_from_file(file_path: str, module_name: str):
    """Import a module from file path."""
    module_parts = module_name.split(".")
    depth = max(0, len(module_parts) - 1)

    # Convert file_path to absolute path if it's not already (note, .resolve will append the current working directory
    # if file_path is relative)
    abs_path = Path(file_path).expanduser().resolve()
    # Ensure depth doesn't exceed available parent directories
    max_available_depth = len(abs_path.parents) - 1

    if max_available_depth < 0:
        # File has no parent directories, use the file's directory itself
        parent_path = str(abs_path.parent)
    else:
        # Clamp depth to available range to avoid IndexError
        depth = min(depth, max_available_depth)
        parent_path = str(abs_path.parents[depth])

    if parent_path not in sys.path:
        sys.path.insert(0, parent_path)

    spec = importlib.util.spec_from_file_location(module_name, file_path)
    if spec is None or spec.loader is None:
        raise ImportError(f"Could not load spec for module {module_name} from {file_path}")

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


#####################################
########## Rsync Helpers ############
#####################################
def generate_rsync_command(subdir: str = ".", exclude_absolute: bool = True):
    """Generate rsync command for syncing from jump pod.

    Args:
        subdir: Directory to sync to (default current directory)
        exclude_absolute: Whether to exclude __absolute__ directory (default True)
    """
    service_name = os.getenv("KT_SERVICE_NAME")
    namespace = os.getenv("POD_NAMESPACE")

    exclude_opt = "--exclude='__absolute__*' " if exclude_absolute else ""
    logger.debug("Syncing code from rsync pod to local directory")
    return f"rsync -av {exclude_opt}rsync://kubetorch-rsync.{namespace}.svc.cluster.local:{RSYNC_PORT}/data/{namespace}/{service_name}/ {subdir}"


def rsync_file_updates():
    """Rsync files from the jump pod to the worker pod.

    Performs two rsync operations in parallel:
    1. Regular files (excluding __absolute__*) to the working directory
    2. Absolute path files (under __absolute__/) to their absolute destinations
    """
    import concurrent.futures
    from concurrent.futures import ThreadPoolExecutor

    service_name = os.getenv("KT_SERVICE_NAME")
    namespace = os.getenv("POD_NAMESPACE")

    # Build base rsync URL
    rsync_base = f"rsync://kubetorch-rsync.{namespace}.svc.cluster.local:{RSYNC_PORT}/data/{namespace}/{service_name}/"

    max_retries = 5
    base_delay = 1  # seconds
    max_delay = 30  # seconds

    def run_rsync_with_retries(rsync_cmd, description):
        """Helper to run rsync with exponential backoff retries."""
        for attempt in range(max_retries):
            resp = subprocess.run(
                rsync_cmd,
                shell=True,
                capture_output=True,
                text=True,
            )

            if resp.returncode == 0:
                logger.debug(f"Successfully rsync'd {description}")
                return  # Success!

            # Check if it's a retryable error
            retryable_errors = [
                "max connections",
                "Temporary failure in name resolution",
                "Name or service not known",
                "Connection refused",
                "No route to host",
            ]

            is_retryable = any(error in resp.stderr for error in retryable_errors)

            if is_retryable and attempt < max_retries - 1:
                # Calculate exponential backoff with jitter
                delay = min(base_delay * (2**attempt) + random.uniform(0, 1), max_delay)
                logger.warning(
                    f"Rsync {description} failed with retryable error: {resp.stderr.strip()}. "
                    f"Retrying in {delay:.1f} seconds (attempt {attempt + 1}/{max_retries})"
                )
                time.sleep(delay)
            else:
                # For non-retryable errors or final attempt, raise immediately
                if attempt == max_retries - 1:
                    logger.error(f"Rsync {description} failed after {max_retries} attempts. Last error: {resp.stderr}")
                raise RuntimeError(f"Rsync {description} failed with error: {resp.stderr}")

        # If we exhausted all retries
        raise RuntimeError(f"Rsync {description} failed after {max_retries} attempts. Last error: {resp.stderr}")

    def rsync_regular_files():
        """Rsync regular files (excluding __absolute__*) to working directory."""
        rsync_cmd_regular = f"rsync -avL --exclude='__absolute__*' {rsync_base} ."
        logger.debug(f"Rsyncing regular files with command: {rsync_cmd_regular}")
        run_rsync_with_retries(rsync_cmd_regular, "regular files")

    def rsync_absolute_files():
        """Rsync absolute path files to their absolute destinations."""
        # First, do a dry-run to see if __absolute__ directory exists
        check_cmd = f"rsync --list-only {rsync_base}__absolute__/"
        check_resp = subprocess.run(check_cmd, shell=True, capture_output=True, text=True)

        if check_resp.returncode == 0 and check_resp.stdout.strip():
            # __absolute__ directory exists, sync its contents to root
            # The trick is to sync from __absolute__/ to / which places files in their absolute paths
            rsync_cmd_absolute = f"rsync -avL {rsync_base}__absolute__/ /"
            logger.debug(f"Rsyncing absolute path files with command: {rsync_cmd_absolute}")
            run_rsync_with_retries(rsync_cmd_absolute, "absolute path files")
        else:
            logger.debug("No absolute path files to sync")

    # Run both rsync operations in parallel
    with ThreadPoolExecutor(max_workers=2) as executor:
        # Submit both tasks
        regular_future = executor.submit(rsync_regular_files)
        absolute_future = executor.submit(rsync_absolute_files)

        # Wait for both to complete and handle any exceptions
        futures = [regular_future, absolute_future]
        for future in concurrent.futures.as_completed(futures):
            try:
                future.result()  # This will raise any exception that occurred
            except Exception as e:
                # Cancel remaining futures if one fails
                for f in futures:
                    f.cancel()
                raise e

    logger.debug("Completed rsync of all files")


#####################################
########### App setup ###############
#####################################
class HealthCheckFilter(logging.Filter):
    def filter(self, record):
        return not (
            isinstance(record.args, tuple)
            and len(record.args) >= 3
            and ("/health" in record.args[2] or record.args[2] == "/")
        )


class RequestContextFilter(logging.Filter):
    def filter(self, record):
        record.request_id = request_id_ctx_var.get("-")
        record.pod = os.getenv("POD_NAME", "unknown-pod")
        return True


class TerminationCheckMiddleware(BaseHTTPMiddleware):
    """Monitor for termination while request is running and return error if detected."""

    async def dispatch(self, request: Request, call_next):
        # Skip health checks and metrics endpoints
        if request.url.path in ["/health", "/", "/metrics"]:
            return await call_next(request)

        # Run the actual request in the background
        import asyncio

        request_task = asyncio.create_task(call_next(request))

        # Monitor for termination while request is running
        while not request_task.done():
            # Check if we're terminating
            if TERMINATION_EVENT.is_set() or (
                hasattr(request.app.state, "terminating") and request.app.state.terminating
            ):
                # Cancel the request task
                request_task.cancel()

                # Return PodTerminatedError
                from kubetorch import PodTerminatedError
                from kubetorch.servers.http.http_server import package_exception

                pod_name = os.environ.get("POD_NAME", "unknown")
                exc = PodTerminatedError(
                    pod_name=pod_name,
                    reason="SIGTERM",
                    status_code=503,
                    events=[
                        {
                            "timestamp": datetime.now(timezone.utc).isoformat(),
                            "reason": "Terminating",
                            "message": "Pod received SIGTERM signal and is shutting down gracefully",
                        }
                    ],
                )

                return package_exception(exc)

            # Wait a bit before checking again or for request to complete
            try:
                result = await asyncio.wait_for(asyncio.shield(request_task), timeout=0.5)
                return result
            except asyncio.TimeoutError:
                # Request still running after 0.5s, continue loop to check termination again
                continue

        # Request completed normally
        return await request_task


class RequestIDMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        request_id = request.headers.get("X-Request-ID", "-")
        token = request_id_ctx_var.set(request_id)

        try:
            response = await call_next(request)
            return response
        finally:
            # Reset the context variable to its default value
            request_id_ctx_var.reset(token)


class StreamToLogger:
    def __init__(self, logger, log_level=logging.INFO, original_stream=None):
        self.logger = logger
        self.log_level = log_level
        self.original_stream = original_stream
        self.linebuf = ""

    def _is_from_logging(self):
        """Check if the current write call is coming from the logging system"""
        frame = sys._getframe()
        while frame:
            if frame.f_globals.get("__name__", "").startswith("logging"):
                return True
            frame = frame.f_back
        return False

    def write(self, buf):
        # Check if this is from logging system
        is_from_logging = self._is_from_logging()

        # Always write to original stream first
        if self.original_stream:
            self.original_stream.write(buf)
            self.original_stream.flush()

        # Skip logging if this is from the logging system to prevent infinite loops
        if self.logger.name == "print_redirect" and is_from_logging:
            return

        # Buffer and log complete lines
        temp_linebuf = self.linebuf + buf
        self.linebuf = ""

        # Split on newlines but keep carriage returns
        lines = []
        current_line = ""
        for char in temp_linebuf:
            if char == "\n":
                lines.append(current_line)
                current_line = ""
            else:
                current_line += char

        # Add any remaining content to linebuf
        if current_line:
            self.linebuf = current_line

        # Log complete lines
        for line in lines:
            if line:
                self.logger.log(self.log_level, line)

    def flush(self):
        if self.original_stream:
            self.original_stream.flush()
        if self.linebuf != "":
            self.logger.log(self.log_level, self.linebuf)
            self.linebuf = ""

    def isatty(self):
        # Delegate to the original stream if it exists, else return False
        if self.original_stream and hasattr(self.original_stream, "isatty"):
            return self.original_stream.isatty()
        return False

    def fileno(self):
        if self.original_stream and hasattr(self.original_stream, "fileno"):
            return self.original_stream.fileno()
        raise OSError("Stream does not support fileno()")

    @property
    def encoding(self):
        # Return the encoding of the original stream if available, else UTF-8
        if self.original_stream and hasattr(self.original_stream, "encoding"):
            return self.original_stream.encoding
        return "utf-8"


# Save original streams before redirection
_original_stdout = sys.stdout
_original_stderr = sys.stderr

# Redirect stdout and stderr to our logger while preserving original streams
sys.stdout = StreamToLogger(print_logger, logging.INFO, _original_stdout)
sys.stderr = StreamToLogger(print_logger, logging.ERROR, _original_stderr)


@asynccontextmanager
async def lifespan(app: FastAPI):
    """Manage application lifecycle"""
    import signal
    import threading

    # Only register signal handlers if we're in the main thread
    # This allows tests to run without signal handling
    if threading.current_thread() is threading.main_thread():
        # Save any existing SIGTERM handler
        original_sigterm_handler = signal.getsignal(signal.SIGTERM)

        def handle_sigterm(signum, frame):
            """Handle SIGTERM for graceful shutdown."""
            logger.info("Received SIGTERM, initiating graceful shutdown...")

            # Mark that we're terminating and interrupt existing requests IMMEDIATELY
            app.state.terminating = True
            TERMINATION_EVENT.set()

            # Clean up distributed supervisor to ensure child processes are terminated
            # This is important because SIGTERM is not propagated to child processes automatically
            # This runs synchronously and may take 1-2 seconds, but existing requests are already interrupted
            global DISTRIBUTED_SUPERVISOR
            if DISTRIBUTED_SUPERVISOR:
                logger.info("Cleaning up distributed supervisor and child processes...")
                try:
                    DISTRIBUTED_SUPERVISOR.cleanup()
                except Exception as e:
                    logger.error(f"Error cleaning up distributed supervisor: {e}")

            # Call the original handler if it exists and isn't the default
            if original_sigterm_handler and original_sigterm_handler not in (
                signal.SIG_DFL,
                signal.SIG_IGN,
            ):
                original_sigterm_handler(signum, frame)

        # Register SIGTERM handler
        signal.signal(signal.SIGTERM, handle_sigterm)
    app.state.terminating = False

    # Startup
    ttl = get_inactivity_ttl_annotation()
    if ttl and KT_OTEL_ENABLED is True:
        app.state.heartbeat_manager = HeartbeatManager(ttl_seconds=ttl)
        if app.state.heartbeat_manager:
            await app.state.heartbeat_manager.start()
            logger.debug(f"Heartbeat manager started with TTL={ttl}s")
    elif ttl:
        logger.warning("TTL annotation found, but OTEL is not enabled, heartbeat disabled")
    else:
        logger.debug("No TTL annotation found, heartbeat disabled")

    try:
        if os.getenv("KT_CALLABLE_TYPE") == "app":
            run_image_setup()
        else:
            load_callable()

        logger.info("Kubetorch Server started.")
        request_id_ctx_var.set("-")  # Reset request_id after launch sequence
        yield

    except Exception:
        # We don't want to raise errors like ImportError during startup, as it will cause the server to crash and the
        # user won't be able to see the error in the logs to debug (e.g. quickly add dependencies or reorganize
        # imports). Instead, we log it (and a stack trace) and continue, so it will be surfaced to the user when they
        # call the service.

        # However if this service is frozen, it should just fail because the user isn't debugging the service and there is no
        # way for the dependencies to be added at runtime.
        logger.error(traceback.format_exc())
        request_id_ctx_var.set("-")
        yield

    finally:
        # Shutdown
        manager = getattr(app.state, "heartbeat_manager", None)
        if manager:
            await manager.stop()
            logger.info("Heartbeat manager stopped")

        # Clean up during normal shutdown so we don't leave any hanging processes, which can cause pods to hang
        # indefinitely. Skip if already cleaned up by SIGTERM handler.
        if DISTRIBUTED_SUPERVISOR and not getattr(app.state, "terminating", False):
            DISTRIBUTED_SUPERVISOR.cleanup()

        # Clear any remaining debugging sessions
        clear_debugging_sessions()


# Add the filter to uvicorn's access logger
logging.getLogger("uvicorn.access").addFilter(HealthCheckFilter())
root_logger = logging.getLogger()
root_logger.addFilter(RequestContextFilter())
for handler in root_logger.handlers:
    handler.addFilter(RequestContextFilter())
print_logger.addFilter(RequestContextFilter())

app = FastAPI(lifespan=lifespan)
app.add_middleware(TerminationCheckMiddleware)  # Check termination first
app.add_middleware(RequestIDMiddleware)

# Configure the FastAPI app for metrics first
# Method will return None for meter_provider if otel is not enabled
app, meter_provider = setup_otel_metrics(app) if KT_OTEL_ENABLED is True else (app, None)

# instrument metrics
if meter_provider is not None:
    try:
        from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor

        logger.info("Instrumenting FastAPI app for metrics only")
        FastAPIInstrumentor.instrument_app(
            app,
            meter_provider=meter_provider,
            excluded_urls="/,/metrics,/health",
        )
    except ImportError:
        logger.info("OpenTelemetry instrumentation not enabled, skipping metrics instrumentation")

# add route for fastapi app
if os.getenv("KT_CALLABLE_TYPE") == "app" and os.getenv("KT_APP_PORT"):
    logger.debug("Adding route for path /http")
    app.add_route(
        "/http/{path:path}",
        _http_reverse_proxy,
        ["GET", "POST", "PUT", "DELETE", "PATCH"],
    )


#####################################
########## Error Handling ###########
#####################################
class ErrorResponse(BaseModel):
    error_type: str
    message: str
    traceback: str
    pod_name: str
    state: Optional[dict] = None  # Optional serialized exception state


# Factor out the exception packaging so we can use it in the handler below and also inside distributed subprocesses
def package_exception(exc: Exception):
    import asyncio
    import concurrent

    error_type = exc.__class__.__name__
    trace = "".join(traceback.format_exception(type(exc), exc, exc.__traceback__))

    # Check if the exception has a status_code attribute (e.g. PodTerminatedError)
    if hasattr(exc, "status_code"):
        status_code = exc.status_code
    elif isinstance(exc, (RequestValidationError, TypeError, AssertionError)):
        status_code = 422
    elif isinstance(exc, (ValueError, UnicodeError, json.JSONDecodeError)):
        status_code = 400
    elif isinstance(exc, (KeyError, FileNotFoundError)):
        status_code = 404
    elif isinstance(exc, PermissionError):
        status_code = 403
    elif isinstance(exc, (StarletteHTTPException, HTTPException)):
        status_code = exc.status_code
    elif isinstance(exc, (MemoryError, OSError)):
        status_code = 500
    elif isinstance(exc, NotImplementedError):
        status_code = 501
    elif isinstance(exc, asyncio.TimeoutError):
        status_code = 504
    elif isinstance(exc, concurrent.futures.TimeoutError):
        status_code = 504
    else:
        status_code = 500

    # Try to serialize exception state if it has __getstate__
    state = None
    if hasattr(exc, "__getstate__"):
        try:
            state = exc.__getstate__()
        except Exception as e:
            logger.debug(f"Could not serialize exception state for {error_type}: {e}")

    error_response = ErrorResponse(
        error_type=error_type,
        message=str(exc),
        traceback=trace,
        pod_name=os.getenv("POD_NAME"),
        state=state,
    )

    return JSONResponse(status_code=status_code, content=error_response.model_dump())


@app.exception_handler(Exception)
async def generic_exception_handler(request: Request, exc: Exception):
    return package_exception(exc)


@app.post("/_reload_image", response_class=JSONResponse)
def _reload_image(
    request: Request,
    deployed_as_of: Optional[str] = Header(None, alias="X-Deployed-As-Of"),
):
    """
    Endpoint to reload the image and metadata configuration.
    This is used to reload the image in cases where we're not calling the callable directly,
    e.g. kt.app and Ray workers.
    """
    global _LAST_DEPLOYED
    deployed_time = (
        datetime.fromisoformat(deployed_as_of).timestamp() if deployed_as_of else datetime.now(timezone.utc).timestamp()
    )
    run_image_setup(deployed_time)
    _LAST_DEPLOYED = deployed_time
    return JSONResponse(
        status_code=200,
        content={"message": "Image and metadata reloaded successfully."},
    )


@app.post("/{cls_or_fn_name}", response_class=JSONResponse)
@app.post("/{cls_or_fn_name}/{method_name}", response_class=JSONResponse)
async def run_callable(
    request: Request,
    cls_or_fn_name: str,
    method_name: Optional[str] = None,
    distributed_subcall=False,
    debug_port: Optional[int] = None,
    params: Optional[Union[Dict, str]] = Body(default=None),
    deployed_as_of: Optional[str] = Header(None, alias="X-Deployed-As-Of"),
    serialization: str = Header("json", alias="X-Serialization"),
):
    if cls_or_fn_name != os.environ["KT_CLS_OR_FN_NAME"]:
        raise HTTPException(
            status_code=404,
            detail=f"Callable '{cls_or_fn_name}' not found in metadata configuration. Found '{os.environ['KT_CLS_OR_FN_NAME']}' instead",
        )

    # NOTE: The distributed replica processes (e.g. PyTorchProcess:run) rely on this running here even though
    # they will reconstruct the callable themselves, because they skip image reloading as a performance optimization.
    # Run load_callable in executor since it may do file I/O and other blocking operations
    callable_obj = await run_in_executor_with_context(None, load_callable, deployed_as_of)

    # If this is a distributed call (and not a subcall from a different distributed replica),
    # and the type of distribution which requires a special call method (e.g. SIMD), use the
    # distributed supervisor to handle the call
    if DISTRIBUTED_SUPERVISOR and DISTRIBUTED_SUPERVISOR.intercept_call():
        # Run the blocking distributed call in executor to avoid blocking the event loop
        result = await run_in_executor_with_context(
            None,
            DISTRIBUTED_SUPERVISOR.call_distributed,
            request,
            cls_or_fn_name,
            method_name,
            params,
            distributed_subcall,
            debug_port,
            deployed_as_of,
        )
        clear_debugging_sessions()
        return result

    # If this is not a distributed call, or the distribution type does not require special handling,
    # run the callable directly
    result = await run_callable_internal(
        callable_obj=callable_obj,
        cls_or_fn_name=cls_or_fn_name,
        method_name=method_name,
        params=params,
        serialization=serialization,
        debug_port=debug_port,
    )
    return result


async def run_callable_internal(
    callable_obj: Callable,
    cls_or_fn_name: str,
    method_name: Optional[str] = None,
    params: Optional[Union[Dict, str]] = Body(default=None),
    serialization: str = "json",
    debug_port: Optional[int] = None,
):
    # Check if serialization is allowed
    allowed_serialization = os.getenv("KT_ALLOWED_SERIALIZATION", DEFAULT_ALLOWED_SERIALIZATION).split(",")
    if serialization not in allowed_serialization:
        raise HTTPException(
            status_code=400,
            detail=f"Serialization format '{serialization}' not allowed. Allowed formats: {allowed_serialization}",
        )

    # Process the call
    args = []
    kwargs = {}
    if params:
        if serialization == "pickle":
            # Handle pickle serialization - extract data from dictionary wrapper
            if isinstance(params, dict) and "data" in params:
                encoded_data = params.pop("data")
                pickled_data = base64.b64decode(encoded_data.encode("utf-8"))
                param_args = pickle.loads(pickled_data)
                # data is unpickled in the format {"args": args, "kwargs": kwargs}
                params.update(param_args)
            elif isinstance(params, str):
                # Fallback for direct string
                pickled_data = base64.b64decode(params.encode("utf-8"))
                params = pickle.loads(pickled_data)

        # Default JSON handling
        args = params.get("args", [])
        kwargs = params.get("kwargs", {})

    if method_name:
        if not hasattr(callable_obj, method_name):
            raise HTTPException(
                status_code=404,
                detail=f"Method '{method_name}' not found in class '{cls_or_fn_name}'",
            )
        user_method = getattr(callable_obj, method_name)
    else:
        user_method = callable_obj

    import inspect

    # Check if the user method is async
    is_async_method = inspect.iscoroutinefunction(user_method)

    callable_name = f"{cls_or_fn_name}.{method_name}" if method_name else cls_or_fn_name
    if debug_port:
        logger.info(f"Debugging remote callable {callable_name} on port {debug_port}")
        deep_breakpoint(debug_port)
        # If using the debugger, step in here ("s") to enter your function/class method.
        if is_async_method:
            result = await user_method(*args, **kwargs)
        else:
            # Run sync method in thread pool to avoid blocking
            # Use lambda to properly pass both args and kwargs
            result = await run_in_executor_with_context(None, lambda: user_method(*args, **kwargs))
    else:
        logger.debug(f"Calling remote callable {callable_name}")
        if is_async_method:
            result = await user_method(*args, **kwargs)
        else:
            # Run sync method in thread pool to avoid blocking
            # Use lambda to properly pass both args and kwargs
            result = await run_in_executor_with_context(None, lambda: user_method(*args, **kwargs))

    # Handle case where sync method returns an awaitable (e.g., from an async framework)
    # This is less common but can happen with some async libraries
    if isinstance(result, Awaitable):
        result = await result

    # Serialize response based on format
    if serialization == "pickle":
        try:
            pickled_result = pickle.dumps(result)
            encoded_result = base64.b64encode(pickled_result).decode("utf-8")
            result = {"data": encoded_result}
        except Exception as e:
            logger.error(f"Failed to pickle result: {str(e)}")
            raise SerializationError(f"Result could not be serialized with pickle: {str(e)}")
    else:
        # Default JSON serialization
        try:
            json.dumps(result)
        except (TypeError, ValueError) as e:
            logger.error(f"Result is not JSON serializable: {str(e)}")
            raise SerializationError(f"Result could not be serialized to JSON: {str(e)}")

    clear_debugging_sessions()

    return result


def run_callable_internal_sync(
    callable_obj: Callable,
    cls_or_fn_name: str,
    method_name: Optional[str] = None,
    params: Optional[Union[Dict, str]] = None,
    serialization: str = "json",
    debug_port: Optional[int] = None,
):
    """Synchronous wrapper for run_callable_internal, used by distributed subprocesses."""
    import asyncio
    import inspect

    # Check if serialization is allowed
    allowed_serialization = os.getenv("KT_ALLOWED_SERIALIZATION", DEFAULT_ALLOWED_SERIALIZATION).split(",")
    if serialization not in allowed_serialization:
        raise HTTPException(
            status_code=400,
            detail=f"Serialization format '{serialization}' not allowed. Allowed formats: {allowed_serialization}",
        )

    # Process the call
    args = []
    kwargs = {}
    if params:
        if serialization == "pickle":
            # Handle pickle serialization - extract data from dictionary wrapper
            if isinstance(params, dict) and "data" in params:
                encoded_data = params.pop("data")
                pickled_data = base64.b64decode(encoded_data.encode("utf-8"))
                param_args = pickle.loads(pickled_data)
                # data is unpickled in the format {"args": args, "kwargs": kwargs}
                params.update(param_args)
            elif isinstance(params, str):
                # Fallback for direct string
                pickled_data = base64.b64decode(params.encode("utf-8"))
                params = pickle.loads(pickled_data)

        # Default JSON handling
        args = params.get("args", [])
        kwargs = params.get("kwargs", {})

    if method_name:
        if not hasattr(callable_obj, method_name):
            raise HTTPException(
                status_code=404,
                detail=f"Method '{method_name}' not found in class '{cls_or_fn_name}'",
            )
        user_method = getattr(callable_obj, method_name)
    else:
        user_method = callable_obj

    # Check if the user method is async
    is_async_method = inspect.iscoroutinefunction(user_method)

    callable_name = f"{cls_or_fn_name}.{method_name}" if method_name else cls_or_fn_name
    if debug_port:
        logger.info(f"Debugging remote callable {callable_name} on port {debug_port}")
        deep_breakpoint(debug_port)
        # If using the debugger, step in here ("s") to enter your function/class method.
        if is_async_method:
            # For async methods in sync context, we need to run them in a new event loop
            result = asyncio.run(user_method(*args, **kwargs))
        else:
            result = user_method(*args, **kwargs)
    else:
        logger.debug(f"Calling remote callable {callable_name}")
        if is_async_method:
            # For async methods in sync context, we need to run them in a new event loop
            result = asyncio.run(user_method(*args, **kwargs))
        else:
            result = user_method(*args, **kwargs)

    # Handle case where sync method returns an awaitable
    if isinstance(result, Awaitable):
        result = asyncio.run(result)

    # Serialize response based on format
    if serialization == "pickle":
        try:
            pickled_result = pickle.dumps(result)
            encoded_result = base64.b64encode(pickled_result).decode("utf-8")
            result = {"data": encoded_result}
        except Exception as e:
            logger.error(f"Failed to pickle result: {str(e)}")
            raise SerializationError(f"Result could not be serialized with pickle: {str(e)}")
    else:
        # Default JSON serialization
        try:
            json.dumps(result)
        except (TypeError, ValueError) as e:
            logger.error(f"Result is not JSON serializable: {str(e)}")
            raise SerializationError(f"Result could not be serialized to JSON: {str(e)}")

    clear_debugging_sessions()

    return result


@app.get("/health", include_in_schema=False)
@app.get("/", include_in_schema=False)
def health():
    return {"status": "healthy"}


if __name__ == "__main__" and not is_running_in_container():
    # NOTE: this will only run in local development, otherwise we start the uvicorn server in the pod template setup
    import uvicorn
    from dotenv import load_dotenv

    load_dotenv()

    logger.info("Starting HTTP server")
    uvicorn.run(app, host="0.0.0.0", port=os.environ.get("KT_SERVER_PORT", 32300))
